# This file is part of the faebryk project
# SPDX-License-Identifier: MIT


import logging
from collections import Counter
from collections.abc import Sequence
from typing import Callable, cast

from faebryk.core.parameter import (
    Abs,
    Add,
    ConstrainableExpression,
    GreaterOrEqual,
    GreaterThan,
    Intersection,
    Is,
    IsSubset,
    Log,
    Multiply,
    Not,
    Or,
    ParameterOperatable,
    Power,
    Round,
    Sin,
    SymmetricDifference,
    Union,
)
from faebryk.core.solver.algorithm import SolverAlgorithm, algorithm
from faebryk.core.solver.mutator import Mutator
from faebryk.core.solver.utils import (
    CanonicalExpression,
    CanonicalNumber,
    Contradiction,
    make_lit,
)
from faebryk.libs.sets.quantity_sets import (
    Quantity_Interval,
)
from faebryk.libs.sets.sets import BoolSet, as_lit
from faebryk.libs.util import cast_assert, groupby, partition_as_list

logger = logging.getLogger(__name__)

# TODO prettify
# - e.g rename from fold


# Boilerplate ==========================================================================

MERGED = False

fold_algorithms: list[SolverAlgorithm] = []
expr_wise_algos: dict[
    type[CanonicalExpression],
    Callable[[CanonicalExpression, Mutator], None],
] = {}


def fold_literals[T: CanonicalExpression](
    mutator: Mutator, expr_type: type[T], f: Callable[[T, Mutator], None]
):
    """
    Tries to do operations on literals or fold expressions.
    - If possible to do literal operation, aliases expr with result.
    - If fold results in new expr, replaces old expr with new one.
    - If fold results in neutralization, returns operand if not literal else alias.

    Examples:
    ```
    Or(True, B) -> alias: True
    Add(A, B, 5, 10) -> replace: Add(A, B, 15)
    Add(10, 15) -> alias: 25
    Not(Not(A)) -> neutralize=replace: A
    ```
    """
    exprs = mutator.nodes_of_type(expr_type, sort_by_depth=True, new_only=False)
    for expr in exprs:
        if mutator.has_been_mutated(expr) or mutator.is_removed(expr):
            continue

        # covered by pure literal folding
        if mutator.utils.is_pure_literal_expression(expr):
            continue

        f(expr, mutator)


@algorithm("Expression-wise", terminal=False)
def expression_wise(mutator: Mutator):
    exprs = mutator.nodes_of_types(
        tuple(expr_wise_algos.keys()), sort_by_depth=True, new_only=False
    )
    exprs_by_type = groupby(exprs, lambda e: type(e))
    for expr_type, exprs in exprs_by_type.items():
        for expr in exprs:
            if mutator.has_been_mutated(expr) or mutator.is_removed(expr):
                continue

            # covered by pure literal folding
            if mutator.utils.is_pure_literal_expression(expr):
                continue
            expr_wise_algos[expr_type](expr, mutator)  # type: ignore


def expression_wise_algorithm[T: CanonicalExpression](expr_type: type[T]):
    def wrap(func: Callable[[T, Mutator], None]) -> SolverAlgorithm:
        if MERGED:
            expr_wise_algos[expr_type] = func  # type: ignore
            if expression_wise not in fold_algorithms:
                fold_algorithms.append(expression_wise)
            return expression_wise
        else:

            @algorithm(f"Fold {expr_type.__name__}", terminal=False)
            def wrapped(mutator: Mutator):
                fold_literals(mutator, expr_type, func)

            fold_algorithms.append(wrapped)
            return wrapped

    return wrap


# TODO REMOVE JUST A TEST
@algorithm("NOOOP")
def noop(mutator: Mutator):
    pass


fold_algorithms.append(noop)


# Arithmetic ---------------------------------------------------------------------------


@expression_wise_algorithm(Add)
def fold_add(expr: Add, mutator: Mutator):
    """
    A + A + 5 + 10 -> 2*A + 15
    A + 5 + (-5) -> A
    A + (-A) + 5 -> Add(5)
    A + (3 * A) + 5 -> (4 * A) + 5
    A + (A * B * 2) -> A + (A * B * 2)
    #TODO think about -> A * (1 + B * 2) (factorization), not in here tho
    A + (B * 2) -> A + (B * 2)
    A + (A * 2) + (A * 3) -> (6 * A)
    A + (A * 2) + ((A *2 ) * 3) -> (3 * A) + (3 * (A * 2))
    #TODO recheck double match (of last case)
    """

    # A + X, B + X, X = A * 5
    # 6*A
    # (A * 2) + (A * 5)
    literal_operands = list(expr.get_operand_literals().values())
    p_operands = expr.get_operand_operatables()
    non_replacable_nonliteral_operands, _replacable_nonliteral_operands = (
        partition_as_list(lambda o: not mutator.has_been_mutated(o), p_operands)
    )
    replacable_nonliteral_operands = Counter(_replacable_nonliteral_operands)
    literal_sum = mutator.utils.fold_op(literal_operands, lambda a, b: a + b, 0)  # type: ignore #TODO

    new_factors, old_factors = mutator.utils.collect_factors(
        replacable_nonliteral_operands, Multiply
    )

    # if non-lit factors all 1 and no literal folding, nothing to do
    if not new_factors and len(literal_sum) == len(literal_operands):
        return

    # Careful, modifying old graph, but should be ok
    factored_operands = [
        mutator.create_expression(Multiply, n, m, from_ops=[expr])
        for n, m in new_factors.items()
    ]

    new_operands = [
        *factored_operands,
        *old_factors,
        *literal_sum,
        *non_replacable_nonliteral_operands,
    ]

    if new_operands == expr.operands:
        return

    # unpack if single operand (operatable)
    if len(new_operands) == 1 and isinstance(new_operands[0], ParameterOperatable):
        new_operands = cast(list[ParameterOperatable], new_operands)
        mutator.mutate_unpack_expression(expr, new_operands)
        return

    new_expr = mutator.mutate_expression(
        expr, operands=new_operands, expression_factory=Add
    )
    # if only one literal operand, equal to it
    if len(new_operands) == 1:
        mutator.utils.alias_to(new_expr, new_operands[0], terminate=True)


@expression_wise_algorithm(Multiply)
def fold_multiply(expr: Multiply, mutator: Mutator):
    """
    TODO doc other simplifications
    A * (A + B)^-1 -> 1 + (B * A^-1)^-1
    """

    literal_operands = list(expr.get_operand_literals().values())
    p_operands = expr.get_operand_operatables()
    non_replacable_nonliteral_operands, _replacable_nonliteral_operands = (
        partition_as_list(lambda o: not mutator.has_been_mutated(o), p_operands)
    )
    replacable_nonliteral_operands = Counter(_replacable_nonliteral_operands)

    literal_prod = mutator.utils.fold_op(literal_operands, lambda a, b: a * b, 1)  # type: ignore #TODO

    new_powers, old_powers = mutator.utils.collect_factors(
        replacable_nonliteral_operands, Power
    )

    # if non-lit powers all 1 and no literal folding, nothing to do
    if not (
        not new_powers
        and len(literal_prod) == len(literal_operands)
        and not (
            literal_prod
            and literal_prod[0] == 0
            and len(replacable_nonliteral_operands)
            + len(non_replacable_nonliteral_operands)
            > 0
        )
    ):
        # Careful, modifying old graph, but should be ok
        powered_operands = [
            mutator.create_expression(Power, n, m, from_ops=[expr])
            for n, m in new_powers.items()
        ]

        new_operands = [
            *powered_operands,
            *old_powers,
            *literal_prod,
            *non_replacable_nonliteral_operands,
        ]

        # 0 * A -> 0
        if 0 in new_operands:
            new_operands = [make_lit(0)]
            # convert_operable_aliased_to_single_into_literal takes care of rest

        # unpack if single operand (operatable)
        if len(new_operands) == 1 and isinstance(new_operands[0], ParameterOperatable):
            new_operands = cast(list[ParameterOperatable], new_operands)
            mutator.mutate_unpack_expression(expr, new_operands)
            return

        if new_operands != expr.operands:
            new_expr = mutator.mutate_expression(
                expr, operands=new_operands, expression_factory=Multiply
            )

            # if only one literal operand, equal to it
            if len(new_operands) == 1:
                mutator.utils.alias_to(new_expr, new_operands[0], terminate=True)
            return

    # if len(expr.operands) == 2:
    #     # TODO pos independend
    #     A, right = expr.operands
    #     # A * (A + B)^-1 -> (1 + B * A^-1)^-1
    #     if (
    #         isinstance(A, ParameterOperatable)
    #         and isinstance(right, Power)
    #         and right.get_operand_literals().get(1) == -1
    #         and isinstance(inner := right.operands[0], Add)
    #         and len(inner.operands) == 2
    #         and not inner.get_operand_literals()
    #         and A in inner.operands
    #     ):
    #         B = inner.get_other_operand(A)
    #         A_inv = mutator.create_expression(
    #             Power,
    #             A,
    #             make_lit(-1),
    #             from_ops=[expr],
    #         )
    #         mutator.mutate_expression(
    #             expr,
    #             expression_factory=Power,
    #             operands=[
    #                 mutator.create_expression(
    #                     Add,
    #                     make_lit(1),
    #                     mutator.create_expression(Multiply, B, A_inv,
    #                       from_ops=[expr]),
    #                 ),
    #                 make_lit(-1),
    #             ],
    #         )


@expression_wise_algorithm(Power)
def fold_pow(expr: Power, mutator: Mutator):
    """
    ```
    A^0 -> 1
    A^1 -> A
    0^A -> 0
    1^A -> 1
    #TODO: (A^B)^C -> A^(B*C)
    #TODO rethink: 0^0 -> 1
    ```
    """

    # TODO if (litex0)^negative -> new constraint

    base, exp = expr.operands

    # All literals
    if mutator.utils.is_numeric_literal(base) and mutator.utils.is_numeric_literal(exp):
        try:
            result = base**exp
        except NotImplementedError:
            # TODO either fix or raise a warning
            return
        mutator.utils.alias_to(expr, result, terminate=True)
        return

    if mutator.utils.is_numeric_literal(exp):
        if exp == 1:
            mutator.mutate_unpack_expression(expr)
            return

        # in python 0**0 is also 1
        if exp == 0:
            mutator.utils.alias_to(expr, as_lit(1), terminate=True)
            return

    if mutator.utils.is_numeric_literal(base):
        if base == 0:
            mutator.utils.alias_to(expr, as_lit(0), terminate=True)
            # FIXME: exp >! 0
            return
        if base == 1:
            mutator.utils.alias_to(expr, as_lit(1), terminate=True)
            return


# @expression_wise_algorithm(Log)
def fold_log(expr: Log, mutator: Mutator):
    """
    # TODO log(A*B) -> log(A) + log(B)
    # TODO log(A^B) -> B*log(A)
    """
    return


# @expression_wise_algorithm(Abs)
def fold_abs(expr: Abs, mutator: Mutator):
    """
    # TODO |-A| = |A|
    # TODO |A*B| = |A|*|B|
    # TODO |A+B| <= |A|+|B|
    """
    return


# @expression_wise_algorithm(Round)
def fold_round(expr: Round, mutator: Mutator):
    """
    TODO: Think about round(A + X)
    """
    return


@expression_wise_algorithm(Sin)
def fold_sin(expr: Sin, mutator: Mutator):
    """
    Sin ss! [-1, 1]
    #TODO Sin(-A) -> -Sin(A)
    #TODO Sin(A + 2*pi) -> Sin(A)
    #TODO Sin(A+B) -> Sin(A)*Cos(B) + Cos(A)*Sin(B)
    """
    mutator.utils.subset_to(expr, make_lit(Quantity_Interval(-1, 1)), from_ops=[expr])


# Setic --------------------------------------------------------------------------------


# @expression_wise_algorithm(Intersection)
def fold_intersect(expr: Intersection, mutator: Mutator):
    """
    Intersection(A) -> A (implicit)
    """

    return


# @expression_wise_algorithm(Union)
def fold_union(expr: Union, mutator: Mutator):
    """
    Union(A) -> A (implicit)
    """

    return


# @expression_wise_algorithm(SymmetricDifference)
def fold_symmetric_difference(expr: SymmetricDifference, mutator: Mutator):
    """ """

    return


# Constrainable ------------------------------------------------------------------------


@expression_wise_algorithm(Or)
def fold_or(expr: Or, mutator: Mutator):
    """
    ```
    Or(A, B, C, True) -> True
    Or(A, B, C, False) -> Or(A, B, C)
    Or(A, B, A) -> Or(A, B)
    Or() -> False
    ```
    """

    # Or(P) -> P implicit (unary identity unpack)
    # Or!(P) -> P! implicit (unary identity unpack)
    # Or(A, B, A) -> Or(A, B) implicit (idempotent)

    # Or(A, B, C, True) -> True
    if BoolSet(True) in expr.get_operand_literals():
        mutator.utils.alias_is_literal_and_check_predicate_eval(expr, True)
        return

    # Or(A, B, C, False) -> Or(A, B, C)
    filtered_operands = [op for op in expr.operands if BoolSet(False) != op]
    if len(filtered_operands) != len(expr.operands):
        # Rebuild without False literals
        mutator.mutate_expression(expr, operands=filtered_operands)
        return


@expression_wise_algorithm(Not)
def fold_not(expr: Not, mutator: Mutator):
    """
    ```
    ¬(¬A) -> A
    ¬P | P constrained -> False

    ¬!(¬A v ¬B v C) -> ¬!(¬!A v ¬!B v C), ¬!C
    ¬!A -> A is! False
    ```
    """
    # TODO ¬(A >= B) -> (B > A) ss ¬(A >= B) (only ss because of partial overlap)

    # ¬(¬A) -> A implicit
    # ¬!(¬A) -> !A implicit

    assert len(expr.operands) == 1
    op = expr.operands[0]
    assert isinstance(op, ParameterOperatable)

    # ¬P | P constrained -> False
    if isinstance(op, ConstrainableExpression) and op.constrained:
        # ¬!P! | P constrained -> Contradiction
        if expr.constrained:
            raise Contradiction("¬!P!", involved=[expr], mutator=mutator)
        mutator.utils.alias_to(expr, as_lit(False), terminate=True)
        return

    if not mutator.has_been_mutated(op):
        # TODO this is kinda ugly, should be in Or fold if it aliases to false
        # ¬!(¬A v ¬B v C) -> ¬!(¬!A v ¬!B v C), ¬!C
        if expr.constrained:
            # ¬( v )
            if isinstance(op, Or):
                # FIXME remove this shortcut
                # should be handle in more general way
                # maybe we need to terminate non-predicates too
                if not op.operands:
                    mutator.utils.alias_is_literal_and_check_predicate_eval(expr, True)
                for inner_op in op.operands:
                    # ¬(¬A v ...)
                    if isinstance(inner_op, Not):
                        for not_op in inner_op.operands:
                            if (
                                isinstance(not_op, ConstrainableExpression)
                                and not not_op.constrained
                            ):
                                mutator.constrain(
                                    cast_assert(
                                        ConstrainableExpression,
                                        mutator.get_copy(not_op),
                                    )
                                )
                    # ¬(A v ...)
                    elif isinstance(inner_op, ConstrainableExpression):
                        parent_nots = inner_op.get_operations(Not)
                        if parent_nots:
                            for n in parent_nots:
                                mutator.constrain(n)
                        else:
                            mutator.create_expression(
                                Not, inner_op, from_ops=[expr], constrain=True
                            )

    if expr.constrained:
        mutator.utils.alias_is_literal_and_check_predicate_eval(op, False)


@expression_wise_algorithm(Is)
def fold_is(expr: Is, mutator: Mutator):
    """
    ```
    P is! True -> P!
    ```
    """

    is_true_alias = expr.constrained and BoolSet(True) in expr.get_operand_literals()
    if is_true_alias:
        # P1 is! True -> P1!
        # P1 is! P2!  -> P1! (implicit)
        for p in expr.get_operand_operatables(ConstrainableExpression):
            mutator.constrain(p)


@expression_wise_algorithm(IsSubset)
def fold_subset(expr: IsSubset, mutator: Mutator):
    """
    ```
    A is B, A ss B | B non(ex)literal -> repr(B, A)
    A ss ([X]) -> A is ([X])
    A ss {} -> A is {}
    # predicates
    P ss! True -> P!
    P ss! False -> ¬!P
    P1 ss! P2! -> P1!
    # literals
    X ss Y -> True / False

    A ss! B, B ss!/is! C -> A ss! C in transitive_subset
    ```
    """

    A, B = expr.operands

    if not mutator.utils.is_literal(B):
        return

    # A ss ([X]) -> A is ([X])
    # A ss {} -> A is {}
    if B.is_single_element() or B.is_empty():
        mutator.mutate_expression(expr, expression_factory=Is)
        return

    if expr.constrained:
        # P1 ss! True -> P1!
        # P1 ss! P2!  -> P1!
        if (
            B == BoolSet(True)
            or isinstance(B, ConstrainableExpression)
            and B.constrained
        ):
            assert isinstance(A, ConstrainableExpression)
            mutator.constrain(A)
        # P ss! False -> ¬!P
        if B == BoolSet(False):
            assert isinstance(A, ConstrainableExpression)
            mutator.create_expression(Not, A, from_ops=[expr], constrain=True)


@expression_wise_algorithm(GreaterOrEqual)
def fold_ge(expr: GreaterOrEqual, mutator: Mutator):
    """
    ```
    A >= A -> True
    # literals
    X >= Y -> [True] / [False] / [True, False]
    A >=! X | |X| > 1 -> A >=! X.max()
    X >=! A | |X| > 1 -> X.min() >=! A
    # uncorrelated
    A >= B{I|X} -> A >= X.max()
    B{I|X} >= A -> X.min() >= A
    ```
    """
    left, right = expr.operands
    literal_operands = cast(Sequence[CanonicalNumber], expr.get_operand_literals())

    # A >=! X | |X| > 1 -> A >=! X.max()
    # X >=! A | |X| > 1 -> X.min() >=! A
    if literal_operands and expr.constrained:
        assert len(literal_operands) == 1
        lit = literal_operands[0]
        if not lit.is_single_element() and not lit.is_empty():
            if left is lit:
                mutator.mutate_expression(
                    expr, operands=[make_lit(lit.min_elem), right]
                )
            else:
                assert right is lit
                mutator.mutate_expression(expr, operands=[left, make_lit(lit.max_elem)])
        return


# @expression_wise_algorithm(GreaterThan)
def fold_gt(expr: GreaterThan, mutator: Mutator):
    """ """
    return



# This file is part of the faebryk project
# SPDX-License-Identifier: MIT

import faebryk.library._F as F


class is_pickable_by_part_number(F.is_pickable.decless()):
    # TODO: make manufacturer an enum
    def __init__(self, manufacturer: str, partno: str):
        super().__init__()
        self._manufacturer = manufacturer
        self._partno = partno

    def get_manufacturer(self) -> str:
        return self._manufacturer

    def get_partno(self) -> str:
        return self._partno



# This file is part of the faebryk project
# SPDX-License-Identifier: MIT


import logging
from typing import cast

from faebryk.core.parameter import (
    ConstrainableExpression,
    Expression,
    IdempotentExpression,
    IdempotentOperands,
    Involutory,
    ParameterOperatable,
    Reflexive,
    UnaryIdentity,
)
from faebryk.core.solver.algorithm import algorithm
from faebryk.core.solver.mutator import Mutator
from faebryk.core.solver.utils import FullyAssociative
from faebryk.libs.util import (
    unique,
)

logger = logging.getLogger(__name__)


@algorithm("Reflexive predicates", terminal=False)
def reflexive_predicates(mutator: Mutator):
    """
    A not lit (done by literal_folding)
    A is A -> True
    A ss A -> True
    A >= A -> True
    """

    predicates = mutator.nodes_of_types(Reflexive, sort_by_depth=True)
    for pred in predicates:
        assert isinstance(pred, ConstrainableExpression)
        if not pred.operatable_operands:
            continue
        if not isinstance(pred.operands[0], ParameterOperatable):
            continue
        if pred.operands[0] is not pred.operands[1]:
            continue

        mutator.utils.alias_is_literal_and_check_predicate_eval(pred, True)


@algorithm("Idempotent deduplicate", terminal=False)
def idempotent_deduplicate(mutator: Mutator):
    """
    Or(A, A, B) -> Or(A, B)
    Union(A, A, B) -> Union(A, B)
    Intersection(A, A, B) -> Intersection(A, B)
    """

    exprs = mutator.nodes_of_types(IdempotentOperands, sort_by_depth=True)
    for expr in exprs:
        assert isinstance(expr, IdempotentOperands)
        unique_operands = unique(expr.operands, key=lambda x: x)
        if len(unique_operands) != len(expr.operands):
            mutator.mutate_expression(expr, operands=unique_operands)


@algorithm("Idempotent unpack", terminal=False)
def idempotent_unpack(mutator: Mutator):
    """
    Abs(Abs(A)) -> Abs(A)
    """

    exprs = mutator.nodes_of_types(IdempotentExpression, sort_by_depth=True)
    for expr in exprs:
        assert isinstance(expr, IdempotentExpression)
        assert len(expr.operands) == 1
        inner = expr.operands[0]
        if type(inner) is not type(expr):
            continue
        mutator.mutate_unpack_expression(expr)


@algorithm("Unary identity unpack", terminal=False)
def unary_identity_unpack(mutator: Mutator):
    """
    E(A), A not lit -> A
    E(A), A lit -> E alias A
    for E in [Add, Multiply, Or, Union, Intersection]
    """

    exprs = mutator.nodes_of_types(UnaryIdentity, sort_by_depth=True)
    for expr in exprs:
        assert isinstance(expr, UnaryIdentity)
        if len(expr.operands) != 1:
            continue
        inner = expr.operands[0]
        if mutator.utils.is_literal(inner):
            mutator.utils.alias_to(expr, inner, terminate=True)
        else:
            mutator.mutate_unpack_expression(expr)


@algorithm("Involutory fold", terminal=False)
def involutory_fold(mutator: Mutator):
    """
    Not(Not(A)) -> A
    """

    exprs = mutator.nodes_of_type(Involutory, sort_by_depth=True)
    for expr in exprs:
        assert isinstance(expr, Involutory)
        if len(expr.operands) != 1:
            continue
        inner = expr.operands[0]
        if type(inner) is not type(expr):
            continue
        assert isinstance(inner, type(expr))
        innest = inner.operands[0]
        if mutator.utils.is_literal(innest):
            mutator.utils.alias_to(expr, innest, terminate=True)
        else:
            mutator.mutator_neutralize_expressions(expr)


@algorithm("Associative expressions", terminal=False)
def associative_flatten(mutator: Mutator):
    """
    Makes
    ```
    (A + B) + (C + (D + E))
       Y    Z    X    W
    -> +(A,B,C,D,E)
       Z'

    for +, *, and, or, &, |, ^
    """
    ops = cast(
        list[FullyAssociative],
        mutator.nodes_of_types(FullyAssociative, sort_by_depth=True),
    )
    # get out deepest expr in compressable tree
    root_ops = [e for e in ops if type(e) not in {type(n) for n in e.get_operations()}]

    def is_replacable(to_replace: Expression, parent_expr: Expression) -> bool:
        """
        Check if an expression can be replaced.
        Only possible if not in use somewhere else or already mapped to new expr
        """
        # overly restrictive: equivalent replacement would be ok
        if mutator.has_been_mutated(to_replace):
            return False
        if to_replace.get_operations() != {parent_expr}:
            return False
        return True

    for expr in root_ops:
        res = mutator.utils.flatten_associative(expr, is_replacable)
        if not res.destroyed_operations:
            continue

        mutator.remove(*res.destroyed_operations)

        mutator.mutate_expression(
            expr,
            operands=res.extracted_operands,
        )



import logging
from pathlib import Path

from mcp.server.fastmcp import FastMCP

from faebryk.libs.util import ConfigFlag

mcp = FastMCP("atopile", stateless_http=True)


logger = logging.getLogger(__name__)

DEBUG = ConfigFlag("MCP_DEBUG", default=False)


def _setup_debug(enable: bool = False):
    """
    Setup debug logging to file
    """
    if not enable:
        return

    handler = logging.FileHandler(Path(__file__).parent / "mcp.log")
    handler.setLevel(logging.DEBUG)
    formatter = logging.Formatter(
        "%(asctime)s - %(name)s - %(levelname)s - %(message)s"
    )
    handler.setFormatter(formatter)
    logger.addHandler(handler)


def run_mcp(http: bool = False, debug: bool = False):
    _setup_debug(enable=bool(DEBUG) or debug)

    from atopile.mcp.tools import (
        cli_tools,
        library_tools,
        packages_tools,
        project_tools,
    )

    cli_tools.install(mcp)
    library_tools.install(mcp)
    packages_tools.install(mcp)
    project_tools.install(mcp)

    logger.info("Starting atopile MCP server...")
    mcp.run(transport="streamable-http" if http else "stdio")



# This file is part of the faebryk project
# SPDX-License-Identifier: MIT
import logging

import faebryk.library._F as F
from faebryk.core.moduleinterface import ModuleInterface
from faebryk.core.parameter import R
from faebryk.libs.library import L
from faebryk.libs.units import P

logger = logging.getLogger(__name__)


class I2S(ModuleInterface):
    sd: F.ElectricLogic  # Serial Data
    ws: F.ElectricLogic  # Word Select (Left/Right Clock)
    sck: F.ElectricLogic  # Serial Clock

    sample_rate = L.p_field(units=P.hertz, domain=R.Domains.Numbers.NATURAL())
    bit_depth = L.p_field(units=P.bit, domain=R.Domains.Numbers.NATURAL())

    @L.rt_field
    def single_electric_reference(self):
        return F.has_single_electric_reference_defined(
            F.ElectricLogic.connect_all_module_references(self)
        )



# This file is part of the faebryk project
# SPDX-License-Identifier: MIT

import faebryk.library._F as F


class has_esphome_config_defined(F.has_esphome_config.impl()):
    def __init__(self, config: dict):
        super().__init__()
        self._config = config

    def get_config(self) -> dict:
        return self._config



# This file is part of the faebryk project
# SPDX-License-Identifier: MIT

import faebryk.library._F as F
from faebryk.core.module import Module
from faebryk.libs.library import L
from faebryk.libs.units import P


class OpAmp(Module):
    bandwidth = L.p_field(units=P.Hz)
    common_mode_rejection_ratio = L.p_field(units=P.dB)
    input_bias_current = L.p_field(units=P.A)
    input_offset_voltage = L.p_field(units=P.V)
    gain_bandwidth_product = L.p_field(units=P.Hz)
    output_current = L.p_field(units=P.A)
    slew_rate = L.p_field(units=P.V / P.s)

    power: F.ElectricPower
    inverting_input: F.Electrical
    non_inverting_input: F.Electrical
    output: F.Electrical

    @L.rt_field
    def simple_value_representation(self):
        S = F.has_simple_value_representation_based_on_params_chain.Spec
        return F.has_simple_value_representation_based_on_params_chain(
            S(self.bandwidth, suffix="BW"),
            S(self.common_mode_rejection_ratio, suffix="CMRR"),
            S(self.input_bias_current, suffix="Ib"),
            S(self.input_offset_voltage, suffix="Vos"),
            S(self.gain_bandwidth_product, suffix="GBW"),
            S(self.output_current, suffix="Iout"),
            S(self.slew_rate, suffix="SR"),
        )

    @L.rt_field
    def pin_association_heuristic(self):
        return F.has_pin_association_heuristic_lookup_table(
            mapping={
                self.power.hv: ["V+", "Vcc", "Vdd", "Vcc+"],
                self.power.lv: ["V-", "Vee", "Vss", "GND", "Vcc-"],
                self.inverting_input: ["-", "IN-"],
                self.non_inverting_input: ["+", "IN+"],
                self.output: ["OUT"],
            },
            accept_prefix=False,
            case_sensitive=False,
        )

    designator_prefix = L.f_field(F.has_designator_prefix)(
        F.has_designator_prefix.Prefix.U
    )

    @property
    def inverting(self) -> F.Electrical:
        return self.inverting_input

    @property
    def non_inverting(self) -> F.Electrical:
        return self.non_inverting_input



from dataclasses import dataclass
from pathlib import Path

from faebryk.libs.kicad.fileformats_latest import C_kicad_pcb_file
from test.end_to_end.conftest import dump_and_run


@dataclass
class PcbSummary:
    num_layers: int
    nets: list[str]
    footprints: list[str]

    @classmethod
    def from_pcb(cls, pcb: C_kicad_pcb_file):
        return cls(
            num_layers=len(pcb.kicad_pcb.layers),
            nets=sorted([net.name for net in pcb.kicad_pcb.nets]),
            footprints=sorted(
                [
                    footprint.propertys.get("Reference").value
                    for footprint in pcb.kicad_pcb.footprints
                ],
            ),
        )

    def add_net(self, net: str) -> "PcbSummary":
        return PcbSummary(
            num_layers=self.num_layers,
            nets=sorted(self.nets + [net]),
            footprints=self.footprints,
        )

    def add_footprint(self, footprint: str) -> "PcbSummary":
        return PcbSummary(
            num_layers=self.num_layers,
            nets=self.nets,
            footprints=sorted(self.footprints + [footprint]),
        )


def summarize_pcb_file(pcb_file: Path) -> PcbSummary:
    pcb = C_kicad_pcb_file.loads(pcb_file.read_text(encoding="utf-8"))
    return PcbSummary.from_pcb(pcb)


SIMPLE_APP = """
import Resistor
module App:
    a = new Resistor
"""

SIMPLE_APP_PCB_SUMMARY = PcbSummary(
    num_layers=29,
    nets=["", "a-net-0", "a-net-1"],
    footprints=["R1"],
)


def test_empty_design(tmpdir: Path):
    pcb_file = tmpdir / Path("layout/app/app.kicad_pcb")
    assert not pcb_file.exists()

    app = """
    module App:
        signal a
    """

    _, stderr, p = dump_and_run(app, [], working_dir=tmpdir)

    assert p.returncode == 0
    assert pcb_file.exists()
    assert "Creating new layout" in stderr

    assert summarize_pcb_file(pcb_file) == PcbSummary(
        num_layers=29, nets=[""], footprints=[]
    )


def test_pcb_file_created(tmpdir: Path):
    pcb_file = tmpdir / Path("layout/app/app.kicad_pcb")
    assert not pcb_file.exists()

    _, stderr, p = dump_and_run(SIMPLE_APP, [], working_dir=tmpdir)

    assert p.returncode == 0
    assert pcb_file.exists()
    assert "Creating new layout" in stderr

    assert SIMPLE_APP_PCB_SUMMARY == summarize_pcb_file(pcb_file)


def test_pcb_file_addition(tmpdir: Path):
    pcb_file = tmpdir / Path("layout/app/app.kicad_pcb")
    assert not pcb_file.exists()

    _, stderr, p = dump_and_run(SIMPLE_APP, [], working_dir=tmpdir)
    assert p.returncode == 0
    assert pcb_file.exists()
    assert "Creating new layout" in stderr
    assert SIMPLE_APP_PCB_SUMMARY == summarize_pcb_file(pcb_file)

    _, stderr, p = dump_and_run(
        f"{SIMPLE_APP}\n    b = new Resistor",
        [],
        working_dir=tmpdir,
    )
    assert p.returncode == 0
    assert "Creating new layout" not in stderr
    assert (
        SIMPLE_APP_PCB_SUMMARY.add_net("b-net-0").add_net("b-net-1").add_footprint("R2")
    ) == summarize_pcb_file(pcb_file)


def test_pcb_file_removal(tmpdir: Path):
    pcb_file = tmpdir / Path("layout/app/app.kicad_pcb")
    assert not pcb_file.exists()

    _, stderr, p = dump_and_run(
        f"{SIMPLE_APP}\n    b = new Resistor",
        [],
        working_dir=tmpdir,
    )
    assert p.returncode == 0
    assert "Creating new layout" in stderr
    assert (
        SIMPLE_APP_PCB_SUMMARY.add_net("b-net-0").add_net("b-net-1").add_footprint("R2")
    ) == summarize_pcb_file(pcb_file)

    _, stderr, p = dump_and_run(SIMPLE_APP, [], working_dir=tmpdir)
    assert p.returncode == 0
    assert "Creating new layout" not in stderr
    assert SIMPLE_APP_PCB_SUMMARY == summarize_pcb_file(pcb_file)



# This file is part of the faebryk project
# SPDX-License-Identifier: MIT

import logging
import math
from bisect import bisect
from collections.abc import Generator
from decimal import Decimal, getcontext
from typing import Any, override

from faebryk.libs.set_math import sine_on_interval
from faebryk.libs.sets.sets import BoolSet, P_Set
from faebryk.libs.util import cast_assert

logger = logging.getLogger(__name__)

REL_DIGITS = 7  # 99.99999% precision
ABS_DIGITS = 15  # femto
EPSILON_REL = 10 ** -(REL_DIGITS - 1)
EPSILON_ABS = 10**-ABS_DIGITS

# To change back to floats, just replace following four lines
getcontext().prec = 30  # quecto/quetta
inf = Decimal(math.inf)
zero = Decimal(0)
Number = Decimal

NumberLike = int | float | Number
NumberLikeR = int, float, Number


def is_int(value: Number) -> bool:
    return value == int(value)


def to_number(value: NumberLike) -> Number:
    if isinstance(value, Number):
        return value
    if isinstance(value, float):
        return Number(float_round(value, ABS_DIGITS))
    return Number(value)


def float_round[T](value: T, digits: int = 0) -> T:
    if not isinstance(value, (float, Number, int)):
        return round(value, digits)  # type: ignore
    if value in [inf, -inf]:
        return value  # type: ignore
    out = round(value, digits)
    return type(value)(out)


def rel_round[T: NumberLike](value: T, digits: int = 0) -> T:
    """ """
    if not isinstance(value, NumberLike):
        raise ValueError("value must be NumberLike")
    if digits < 0:
        raise ValueError("digits must be non-negative")

    if value in [inf, -inf]:
        return value  # type: ignore
    if value == 0:
        return value  # type: ignore

    a_val = abs(value)
    if a_val >= 1:
        magnitude = math.floor(math.log10(a_val)) + 1
        digits -= magnitude

    out = round(value, digits)
    return type(value)(out)


def eq(value: NumberLike, other: NumberLike) -> bool:
    return math.isclose(value, other, rel_tol=EPSILON_REL, abs_tol=EPSILON_ABS)


def ge(value: NumberLike, other: NumberLike) -> bool:
    return value >= other or eq(value, other)


# Numeric ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
class Numeric_Set(P_Set):
    pass


class Numeric_Interval(Numeric_Set):
    """
    Numeric interval [min, max].
    """

    def __init__(self, min: NumberLike, max: NumberLike):
        self._min = to_number(min)
        self._max = to_number(max)

        if not self._min <= self._max:
            raise ValueError("min must be less than or equal to max")
        if self._min == inf or self._max == -inf:
            raise ValueError("min or max has bad infinite value")

        super().__init__()

    def is_empty(self) -> bool:
        return False

    def is_unbounded(self) -> bool:
        return self._min == -inf and self._max == inf

    @override
    def is_finite(self) -> bool:
        return self._min != -inf and self._max != inf

    @property
    def min_elem(self) -> Number:
        return self._min

    @property
    def max_elem(self) -> Number:
        return self._max

    @property
    def is_integer(self) -> bool:
        return self.is_single_element() and is_int(self._min)

    def as_center_rel(self) -> tuple[Number, Number]:
        if self._min == self._max:
            return self._min, zero
        if not self.is_finite():
            return self.min_elem, inf

        center = (self._min + self._max) / 2
        if center == 0:
            rel = inf
        else:
            rel = (self._max - self._min) / 2 / center
        rel = abs(rel)
        return center, rel  # type: ignore

    def is_subset_of(self, other: "Numeric_Interval") -> bool:
        return ge(self._min, other._min) and ge(other._max, self._max)

    def op_add_interval(self, other: "Numeric_Interval") -> "Numeric_Interval":
        """
        Arithmetically adds two intervals.
        """
        return Numeric_Interval(self._min + other._min, self._max + other._max)

    def op_negate(self) -> "Numeric_Interval":
        """
        Arithmetically negates a interval.
        """
        return Numeric_Interval(-self._max, -self._min)

    def op_subtract_interval(self, other: "Numeric_Interval") -> "Numeric_Interval":
        """
        Arithmetically subtracts a interval from another interval.
        """
        return self.op_add_interval(other.op_negate())

    def op_mul_interval(self, other: "Numeric_Interval") -> "Numeric_Interval":
        """
        Arithmetically multiplies two intervals.
        """

        # TODO decide on definition
        # def guarded_mul(a: Number, b: Number) -> list:
        #     """
        #     0.0 * inf -> [0.0, inf]
        #     0.0 * -inf -> [-inf, 0.0]
        #     """
        #     if 0.0 in [a, b] and inf in [a, b]:
        #         assert isinstance(a, Number) or isinstance(b, Number)
        #         return [0.0, inf]
        #     if 0.0 in [a, b] and -inf in [a, b]:
        #         assert isinstance(a, Number) or isinstance(b, Number)
        #         return [0.0, -inf]
        #     prod = a * b
        #     assert not math.isnan(prod)
        #     return [prod]

        def guarded_mul(a: Number, b: Number) -> list:
            """
            0 * inf -> 0
            0 * -inf -> 0
            """
            if 0.0 in [a, b]:
                return [0.0]  # type: ignore
            prod = a * b
            assert not math.isnan(prod)
            return [prod]

        values = [
            res
            for a, b in [
                (self._min, other._min),
                (self._min, other._max),
                (self._max, other._min),
                (self._max, other._max),
            ]
            for res in guarded_mul(a, b)
        ]
        _min = min(values)
        _max = max(values)

        return Numeric_Interval(_min, _max)

    def op_pow_interval(self, other: "Numeric_Interval") -> "Numeric_Interval_Disjoint":
        base = self
        exp = other

        if exp.max_elem < 0:
            return base.op_pow_interval(exp.op_negate()).op_invert()
        if exp.min_elem < 0:
            raise NotImplementedError("crossing zero in exp not implemented yet")
        if base.min_elem < 0 and not exp.is_integer:
            raise NotImplementedError(
                "cannot raise negative base to fractional exponent (complex result)"
            )

        def _pow(x, y):
            try:
                return x**y
            except OverflowError:
                return inf if x > 0 else -inf

        a, b = base._min, base._max
        c, d = exp._min, exp._max

        # see first two guards above
        assert c >= 0

        values = [
            _pow(a, c),
            _pow(a, d),
            _pow(b, c),
            _pow(b, d),
        ]

        if a < 0 < b:
            # might be 0 exp, so just in case applying exponent
            values.extend((zero**c, zero**d))

            # d odd
            if d % 2 == 1:
                # c < k < d
                if (k := d - 1) > c:
                    values.append(_pow(a, k))

        return Numeric_Interval_Disjoint(Numeric_Interval(min(values), max(values)))

    def op_invert(self) -> "Numeric_Interval_Disjoint":
        """
        Arithmetically inverts a interval (1/x).
        """
        if self._min == 0 == self._max:
            return Numeric_Set_Empty()
        if self._min < 0 < self._max:
            return Numeric_Interval_Disjoint(
                Numeric_Interval(-inf, 1 / self._min),
                Numeric_Interval(1 / self._max, inf),
            )
        elif self._min < 0 == self._max:
            return Numeric_Interval_Disjoint(Numeric_Interval(-inf, 1 / self._min))
        elif self._min == 0 < self._max:
            return Numeric_Interval_Disjoint(Numeric_Interval(1 / self._max, inf))
        else:
            return Numeric_Interval_Disjoint(
                Numeric_Interval(1 / self._max, 1 / self._min)
            )

    def op_div_interval(
        self: "Numeric_Interval", other: "Numeric_Interval"
    ) -> "Numeric_Interval_Disjoint":
        """
        Arithmetically divides a interval by another interval.
        """
        # TODO not sure I like this
        # this is very numerically unstable
        # [0] / [0, 1] ->  [0]
        # [1e-20] / [0, 1] -> [1e-20, inf]
        # if self.is_single_element() and self._min == 0:
        #     return Numeric_Interval_Disjoint(
        #         Numeric_Interval(self._min, self._max),
        #     )
        return Numeric_Interval_Disjoint(
            *(self.op_mul_interval(o) for o in other.op_invert().intervals)
        )

    def op_intersect_interval(
        self, other: "Numeric_Interval"
    ) -> "Numeric_Interval_Disjoint":
        """
        Set intersects two intervals.
        """
        min_ = max(self._min, other._min)
        max_ = min(self._max, other._max)
        if min_ <= max_:
            return Numeric_Interval_Disjoint(Numeric_Interval(min_, max_))
        if eq(min_, max_):
            return Numeric_Set_Discrete(min_)
        return Numeric_Set_Empty()

    def op_difference_interval(
        self, other: "Numeric_Interval"
    ) -> "Numeric_Interval_Disjoint":
        """
        Set difference of two intervals.
        """
        # TODO feels like there is more pretty way to implement this function

        # no overlap
        if self._max < other._min or self._min > other._max:
            return Numeric_Interval_Disjoint(self)
        # fully covered
        if other._min <= self._min and other._max >= self._max:
            return Numeric_Set_Empty()
        # inner overlap
        if self._min < other._min and self._max > other._max:
            return Numeric_Interval_Disjoint(
                Numeric_Interval(self._min, other._min),
                Numeric_Interval(other._max, self._max),
            )
        # right overlap
        if self._min < other._min:
            return Numeric_Interval_Disjoint(Numeric_Interval(self._min, other._min))
        # left overlap
        return Numeric_Interval_Disjoint(Numeric_Interval(other._max, self._max))

    def op_round(self, ndigits: int = 0) -> "Numeric_Interval":
        return Numeric_Interval(
            float_round(self._min, ndigits), float_round(self._max, ndigits)
        )  # type: ignore #TODO

    def op_abs(self) -> "Numeric_Interval":
        if self._min < 0 < self._max:
            return Numeric_Interval(0, self._max)  # type: ignore #TODO
        if self._min < 0 and self._max < 0:
            return Numeric_Interval(-self._max, -self._min)
        if self._min < 0 and self._max == 0:
            return Numeric_Interval(0, -self._min)
        if self._min == 0 and self._max < 0:
            return Numeric_Interval(self._max, 0)

        assert self._min >= 0 and self._max >= 0
        return self

    def op_log(self) -> "Numeric_Interval":
        if self._min <= 0:
            raise ValueError(f"invalid log of {self}")
        return Numeric_Interval(math.log(self._min), math.log(self._max))  # type: ignore #TODO

    def op_sin(self) -> "Numeric_Interval":
        return Numeric_Interval(*sine_on_interval((float(self._min), float(self._max))))

    def maybe_merge_interval(
        self, other: "Numeric_Interval"
    ) -> list["Numeric_Interval"]:
        """
        Attempts to merge two intervals if they overlap or are adjacent.

        Example:
            - [1,5] and [3,7] merge to [1,7] since 3 falls within [1,5]
            - [1,2] and [4,5] stay separate since 4 doesn't fall within [1,2]

        Returns:
            List containing either:
            - Single merged interval if intervals overlap
            - Both intervals in order if they don't overlap
        """
        is_left = self._min <= other._min
        left = self if is_left else other
        right = other if is_left else self
        if right._min in self:
            return [Numeric_Interval(left._min, max(left._max, right._max))]
        return [left, right]

    def __eq__(self, other: Any) -> bool:
        """
        Set checks if two intervals are equal.
        """
        if not isinstance(other, Numeric_Interval):
            return False

        return eq(self._min, other._min) and eq(self._max, other._max)

    def __contains__(self, item: Any) -> bool:
        """
        Set checks if a number is in a interval.
        """
        if not isinstance(item, NumberLike):
            return False
        return ge(self._max, item) and ge(item, self._min)

    def __hash__(self) -> int:
        return hash((self._min, self._max))

    def __repr__(self) -> str:
        return f"_interval({self._min}, {self._max})"

    def __str__(self) -> str:
        if self._min == self._max:
            return f"[{self._min}]"
        center, rel = self.as_center_rel()
        if rel < 1:
            return f"{center} ± {rel * 100}%"
        return f"[{self._min}, {self._max}]"

    def __add__(self, other: "Numeric_Interval"):
        return self.op_add_interval(other)

    def __sub__(self, other: "Numeric_Interval"):
        return self.op_subtract_interval(other)

    def __neg__(self):
        return self.op_negate()

    def __mul__(self, other: "Numeric_Interval"):
        return self.op_mul_interval(other)

    def __truediv__(self: "Numeric_Interval", other: "Numeric_Interval"):
        return self.op_div_interval(other)

    def __and__(self, other: "Numeric_Interval"):
        return self.op_intersect_interval(other)

    def __or__(self, other: "Numeric_Interval"):
        return self.maybe_merge_interval(other)

    def __pow__(self, other: "Numeric_Interval"):
        return self.op_pow_interval(other)

    def is_single_element(self) -> bool:
        return eq(self._min, self._max)

    @override
    def any(self) -> Number:
        return self._min

    @override
    def serialize_pset(self) -> dict:
        return {
            "min": None if math.isinf(self._min) else float(self._min),
            "max": None if math.isinf(self._max) else float(self._max),
        }

    @override
    @classmethod
    def deserialize_pset(cls, data: dict):
        min_ = data["min"] if data["min"] is not None else -inf
        max_ = data["max"] if data["max"] is not None else inf
        return cls(min_, max_)  # type: ignore


def Numeric_Singleton(value: Number) -> Numeric_Interval:
    """
    Set containing a single value. \n
    Represented by a Numeric_Interval with min and max being the same value.
    """
    return Numeric_Interval(value, value)


class Numeric_Interval_Disjoint(Numeric_Set):
    """
    Numeric Interval (min < max) with gaps. \n
    Represented by Set of multiple continuous Numeric Intervals.
    """

    def __init__(
        self,
        *intervals: "Numeric_Interval | Numeric_Interval_Disjoint",
    ):
        def gen_flat_non_empty() -> Generator[Numeric_Interval]:
            for r in intervals:
                if r.is_empty():
                    continue
                if isinstance(r, Numeric_Interval_Disjoint):
                    yield from r.intervals
                else:
                    assert isinstance(r, Numeric_Interval)
                    yield r

        non_empty_intervals = list(gen_flat_non_empty())
        sorted_intervals = sorted(non_empty_intervals, key=lambda e: e.min_elem)

        def gen_merge():
            last = None
            for interval in sorted_intervals:
                if last is None:
                    last = interval
                else:
                    *prefix, last = last.maybe_merge_interval(interval)
                    yield from prefix
            if last is not None:
                yield last

        self.intervals = list(gen_merge())

    def is_empty(self) -> bool:
        return len(self.intervals) == 0

    def is_unbounded(self) -> bool:
        if self.is_empty():
            return False
        return self.intervals[0].is_unbounded()

    @override
    def is_finite(self) -> bool:
        if self.is_empty():
            return True
        return self.intervals[0].is_finite() and self.intervals[-1].is_finite()

    @property
    def min_elem(self) -> Number:
        if self.is_empty():
            raise ValueError("empty interval cannot have min element")
        return self.intervals[0].min_elem

    @property
    def max_elem(self) -> Number:
        if self.is_empty():
            raise ValueError("empty interval cannot have max element")
        return self.intervals[-1].max_elem

    def closest_elem(self, target: NumberLike) -> Number:
        assert isinstance(target, NumberLike)
        target = to_number(target)
        if self.is_empty():
            raise ValueError("empty interval cannot have closest element")
        index = bisect(self.intervals, target, key=lambda r: r.min_elem)
        left = self.intervals[index - 1] if index > 0 else None
        if left is not None and target in left:
            return target
        left_bound = left.max_elem if left is not None else None
        right_bound = (
            self.intervals[index].min_elem if index < len(self.intervals) else None
        )
        try:
            [one] = [b for b in [left_bound, right_bound] if b is not None]
            return one
        except ValueError:
            assert left_bound and right_bound
            if target - left_bound < right_bound - target:
                return left_bound
            return right_bound
        assert False  # unreachable

    def is_superset_of(self, other: "Numeric_Interval_Disjoint") -> bool:
        return other == other.op_intersect_intervals(self)

    def is_subset_of(self, other: "Numeric_Interval_Disjoint") -> bool:
        return other.is_superset_of(self)

    def op_intersect_interval(
        self, other: "Numeric_Interval"
    ) -> "Numeric_Interval_Disjoint":
        return Numeric_Interval_Disjoint(
            *(r.op_intersect_interval(other) for r in self.intervals)
        )

    def op_intersect_intervals(
        self, other: "Numeric_Interval_Disjoint"
    ) -> "Numeric_Interval_Disjoint":
        result = []
        s, o = 0, 0
        while s < len(self.intervals) and o < len(other.intervals):
            rs, ro = self.intervals[s], other.intervals[o]
            intersect = rs.op_intersect_interval(ro)
            if not intersect.is_empty():
                result.append(intersect)

            if rs.max_elem < ro.min_elem:
                # no remaining element in other list can intersect with rs
                s += 1
            elif ro.max_elem < rs.min_elem:
                # no remaining element in self list can intersect with ro
                o += 1
            elif rs.max_elem < ro.max_elem:
                # rs ends before ro, so move to next in self list
                s += 1
            elif ro.max_elem < rs.max_elem:
                # ro ends before rs, so move to next in other list
                o += 1
            else:
                # rs and ro end on approximately same number, move to next in both lists
                s += 1
                o += 1

        return Numeric_Interval_Disjoint(*result)

    def op_union_intervals(
        self, other: "Numeric_Interval_Disjoint"
    ) -> "Numeric_Interval_Disjoint":
        return Numeric_Interval_Disjoint(*self.intervals, *other.intervals)

    def op_difference_interval(
        self, other: "Numeric_Interval"
    ) -> "Numeric_Interval_Disjoint":
        return Numeric_Interval_Disjoint(
            *(r.op_difference_interval(other) for r in self.intervals)
        )

    def op_difference_intervals(
        self, other: "Numeric_Interval_Disjoint"
    ) -> "Numeric_Interval_Disjoint":
        # TODO there is probably a more efficient way to do this
        out = self
        for o in other.intervals:
            out = out.op_difference_interval(o)
        return out

    def op_symmetric_difference_intervals(
        self, other: "Numeric_Interval_Disjoint"
    ) -> "Numeric_Interval_Disjoint":
        return self.op_union_intervals(other).op_difference_intervals(
            self.op_intersect_intervals(other)
        )

    def op_add_intervals(
        self, other: "Numeric_Interval_Disjoint"
    ) -> "Numeric_Interval_Disjoint":
        return Numeric_Interval_Disjoint(
            *(r.op_add_interval(o) for r in self.intervals for o in other.intervals)
        )

    def op_negate(self) -> "Numeric_Interval_Disjoint":
        return Numeric_Interval_Disjoint(*(r.op_negate() for r in self.intervals))

    def op_subtract_intervals(
        self, other: "Numeric_Interval_Disjoint"
    ) -> "Numeric_Interval_Disjoint":
        return self.op_add_intervals(other.op_negate())

    def op_mul_intervals(
        self, other: "Numeric_Interval_Disjoint"
    ) -> "Numeric_Interval_Disjoint":
        return Numeric_Interval_Disjoint(
            *(r.op_mul_interval(o) for r in self.intervals for o in other.intervals)
        )

    def op_invert(self) -> "Numeric_Interval_Disjoint":
        return Numeric_Interval_Disjoint(*(r.op_invert() for r in self.intervals))

    def op_div_intervals(
        self: "Numeric_Interval_Disjoint",
        other: "Numeric_Interval_Disjoint",
    ) -> "Numeric_Interval_Disjoint":
        return self.op_mul_intervals(other.op_invert())

    def op_pow_intervals(
        self, other: "Numeric_Interval_Disjoint"
    ) -> "Numeric_Interval_Disjoint":
        return Numeric_Interval_Disjoint(
            *(r.op_pow_interval(o) for r in self.intervals for o in other.intervals)
        )

    def op_ge_intervals(self, other: "Numeric_Interval_Disjoint") -> BoolSet:
        if self.is_empty() or other.is_empty():
            return BoolSet()
        if self.min_elem >= other.max_elem:
            return BoolSet(True)
        if self.max_elem < other.min_elem:
            return BoolSet(False)
        return BoolSet(True, False)

    def op_gt_intervals(self, other: "Numeric_Interval_Disjoint") -> BoolSet:
        if self.is_empty() or other.is_empty():
            return BoolSet()
        if self.min_elem > other.max_elem:
            return BoolSet(True)
        if self.max_elem <= other.min_elem:
            return BoolSet(False)
        return BoolSet(True, False)

    def op_le_intervals(self, other: "Numeric_Interval_Disjoint") -> BoolSet:
        if self.is_empty() or other.is_empty():
            return BoolSet()
        if self.max_elem <= other.min_elem:
            return BoolSet(True)
        if self.min_elem > other.max_elem:
            return BoolSet(False)
        return BoolSet(True, False)

    def op_lt_intervals(self, other: "Numeric_Interval_Disjoint") -> BoolSet:
        if self.is_empty() or other.is_empty():
            return BoolSet()
        if self.max_elem < other.min_elem:
            return BoolSet(True)
        if self.min_elem >= other.max_elem:
            return BoolSet(False)
        return BoolSet(True, False)

    def op_round(self, ndigits: int = 0) -> "Numeric_Interval_Disjoint":
        return Numeric_Interval_Disjoint(*(r.op_round(ndigits) for r in self.intervals))

    def op_abs(self) -> "Numeric_Interval_Disjoint":
        return Numeric_Interval_Disjoint(*(r.op_abs() for r in self.intervals))

    def op_log(self) -> "Numeric_Interval_Disjoint":
        return Numeric_Interval_Disjoint(*(r.op_log() for r in self.intervals))

    def op_sin(self) -> "Numeric_Interval_Disjoint":
        return Numeric_Interval_Disjoint(*(r.op_sin() for r in self.intervals))

    def __contains__(self, item: Any) -> bool:
        if not isinstance(item, NumberLike):
            return False
        index = bisect(self.intervals, item, key=lambda r: r.min_elem)

        if index == 0:
            return False
        return item in self.intervals[index - 1]

    def __eq__(self, value: Any) -> bool:
        if not isinstance(value, Numeric_Interval_Disjoint):
            return False
        if len(self.intervals) != len(value.intervals):
            return False
        for r1, r2 in zip(self.intervals, value.intervals):
            if r1 != r2:
                return False
        return True

    def __ge__(self, other: "Numeric_Interval_Disjoint") -> BoolSet:
        return self.op_ge_intervals(other)

    def __gt__(self, other: "Numeric_Interval_Disjoint") -> BoolSet:
        return self.op_gt_intervals(other)

    def __le__(self, other: "Numeric_Interval_Disjoint") -> BoolSet:
        return self.op_le_intervals(other)

    def __lt__(self, other: "Numeric_Interval_Disjoint") -> BoolSet:
        return self.op_lt_intervals(other)

    def __hash__(self) -> int:
        return hash(tuple(hash(r) for r in self.intervals))

    def __repr__(self) -> str:
        return f"_N_intervals({
            ', '.join(f'[{r._min}, {r._max}]' for r in self.intervals)
        })"

    def __iter__(self) -> Generator["Numeric_Interval"]:
        yield from self.intervals

    # operators
    def __add__(self, other: "Numeric_Interval_Disjoint"):
        return self.op_add_intervals(other)

    def __sub__(self, other: "Numeric_Interval_Disjoint"):
        return self.op_subtract_intervals(other)

    def __neg__(self):
        return self.op_negate()

    def __mul__(self, other: "Numeric_Interval_Disjoint"):
        return self.op_mul_intervals(other)

    def __truediv__(
        self: "Numeric_Interval_Disjoint",
        other: "Numeric_Interval_Disjoint",
    ):
        return self.op_div_intervals(other)

    def __and__(
        self,
        other: "Numeric_Interval_Disjoint | Numeric_Interval",
    ):
        if isinstance(other, Numeric_Interval):
            return self.op_intersect_interval(other)
        return self.op_intersect_intervals(other)

    def __or__(self, other: "Numeric_Interval_Disjoint"):
        return self.op_union_intervals(other)

    def __pow__(self, other: "Numeric_Interval_Disjoint"):
        return self.op_pow_intervals(other)

    def is_single_element(self) -> bool:
        if self.is_empty():
            return False
        return self.min_elem == self.max_elem

    @override
    def any(self) -> Number:
        return self.min_elem

    @override
    def serialize_pset(self) -> dict:
        return {"intervals": [r.serialize() for r in self.intervals]}

    @override
    @classmethod
    def deserialize_pset(cls, data: dict):
        return cls(
            *[
                cast_assert(Numeric_Interval, P_Set.deserialize(r))
                for r in data["intervals"]
            ]
        )


class Numeric_Set_Discrete(Numeric_Interval_Disjoint):
    """
    Numeric Set of multiple single values. \n
    Represented by Set of Numeric Singletons
    (each being an Interval with single value).
    """

    def __init__(self, *values: Number):
        super().__init__(*(Numeric_Singleton(v) for v in values))

    def iter_singles(self) -> Generator[Number]:
        for r in self.intervals:
            yield r._min


def Numeric_Set_Empty() -> Numeric_Interval_Disjoint:
    """
    Empty Numeric Set.
    Represented by an empty Numeric_Interval_Disjoint.
    """
    return Numeric_Interval_Disjoint()



# This file is part of the faebryk project
# SPDX-License-Identifier: MIT

from enum import Enum, auto


class SMDSize(Enum):
    class UnableToConvert(Exception): ...

    I01005 = auto()
    I0201 = auto()
    I0402 = auto()
    I0603 = auto()
    I0805 = auto()
    I1206 = auto()
    I1210 = auto()
    I1808 = auto()
    I1812 = auto()
    I1825 = auto()
    I2220 = auto()
    I2225 = auto()
    I3640 = auto()

    M0402 = auto()
    M0603 = auto()
    M1005 = auto()
    M1608 = auto()
    M2012 = auto()
    M3216 = auto()
    M3225 = auto()
    M4520 = auto()
    M4532 = auto()
    M4564 = auto()
    M5750 = auto()
    M5664 = auto()
    M9110 = auto()

    SMD4x4mm = "SMD,4x4mm"
    SMD6x6mm = "SMD,6x6mm"
    SMD5x5mm = "SMD,5x5mm"
    SMD3x3mm = "SMD,3x3mm"
    SMD8x8mm = "SMD,8x8mm"
    SMD12x12mm = "SMD,12x12mm"
    SMD12_5x12_5mm = "SMD,12.5x12.5mm"
    SMD7_8x7mm = "SMD,7.8x7mm"
    SMD4_5x4mm = "SMD,4.5x4mm"
    SMD11_5x10mm = "SMD,11.5x10mm"
    SMD6_6x7mm = "SMD,6.6x7mm"
    SMD7x6_6mm = "SMD,7x6.6mm"
    SMD5_8x5_2mm = "SMD,5.8x5.2mm"
    SMD6_6x7_3mm = "SMD,6.6x7.3mm"
    SMD3_5x3mm = "SMD,3.5x3mm"
    SMD7_3x7_3mm = "SMD,7.3x7.3mm"
    SMD6_6x7_1mm = "SMD,6.6x7.1mm"
    SMD7x7mm = "SMD,7x7mm"
    SMD5_4x5_2mm = "SMD,5.4x5.2mm"
    SMD6_7x6_7mm = "SMD,6.7x6.7mm"
    SMD11x10mm = "SMD,11x10mm"
    SMD10x11mm = "SMD,10x11mm"
    SMD5_2x5_8mm = "SMD,5.2x5.8mm"
    SMD4_4x4_2mm = "SMD,4.4x4.2mm"
    SMD13_8x12_6mm = "SMD,13.8x12.6mm"
    SMD10_1x10_1mm = "SMD,10.1x10.1mm"
    SMD13_5x12_6mm = "SMD,13.5x12.6mm"
    SMD4_7x4_7mm = "SMD,4.7x4.7mm"
    SMD12_3x12_3mm = "SMD,12.3x12.3mm"
    SMD12_6x13_5mm = "SMD,12.6x13.5mm"
    SMD2_8x2_9mm = "SMD,2.8x2.9mm"
    SMD7_3x6_6mm = "SMD,7.3x6.6mm"
    SMD2_5x2mm = "SMD,2.5x2mm"
    SMD4_9x4_9mm = "SMD,4.9x4.9mm"
    SMD10_2x10mm = "SMD,10.2x10mm"
    SMD7_1x6_6mm = "SMD,7.1x6.6mm"
    SMD10x10mm = "SMD,10x10mm"
    SMD5_7x5_7mm = "SMD,5.7x5.7mm"
    SMD4_1x4_1mm = "SMD,4.1x4.1mm"
    SMD4_1x4_5mm = "SMD,4.1x4.5mm"
    SMD7x7_8mm = "SMD,7x7.8mm"
    SMD10x9mm = "SMD,10x9mm"
    SMD0_6x1_2mm = "SMD,0.6x1.2mm"
    SMD6_5x6_9mm = "SMD,6.5x6.9mm"
    SMD1_6x2mm = "SMD,1.6x2mm"
    SMD2x2_5mm = "SMD,2x2.5mm"
    SMD7_1x6_5mm = "SMD,7.1x6.5mm"
    SMD8x8_5mm = "SMD,8x8.5mm"
    SMD4_5x4_1mm = "SMD,4.5x4.1mm"
    SMD4_2x4_4mm = "SMD,4.2x4.4mm"
    SMD10_4x10_3mm = "SMD,10.4x10.3mm"
    SMD10x11_5mm = "SMD,10x11.5mm"
    SMD13_5x12_8mm = "SMD,13.5x12.8mm"
    SMD17_2x17_2mm = "SMD,17.2x17.2mm"
    SMD5_2x5_4mm = "SMD,5.2x5.4mm"
    SMD11_6x10_1mm = "SMD,11.6x10.1mm"
    SMD10_5x10_3mm = "SMD,10.5x10.3mm"
    SMD7_2x6_6mm = "SMD,7.2x6.6mm"
    SMD10x10_2mm = "SMD,10x10.2mm"
    SMD7_8x7_8mm = "SMD,7.8x7.8mm"
    SMD1_7x2_3mm = "SMD,1.7x2.3mm"
    SMD5_2x5_7mm = "SMD,5.2x5.7mm"
    SMD2x2mm = "SMD,2x2mm"
    SMD4_5x5_2mm = "SMD,4.5x5.2mm"
    SMD9x10mm = "SMD,9x10mm"
    SMD2_5x2_9mm = "SMD,2.5x2.9mm"
    SMD4_6x4_1mm = "SMD,4.6x4.1mm"
    SMD7_5x7_5mm = "SMD,7.5x7.5mm"
    SMD5_5x5_2mm = "SMD,5.5x5.2mm"
    SMD6_4x6_6mm = "SMD,6.4x6.6mm"
    SMD12_5x13_5mm = "SMD,12.5x13.5mm"
    SMD10_7x10mm = "SMD,10.7x10mm"
    SMD5_5x5_3mm = "SMD,5.5x5.3mm"
    SMD10_1x11_6mm = "SMD,10.1x11.6mm"
    SMD10_3x10_5mm = "SMD,10.3x10.5mm"
    SMD3_2x3mm = "SMD,3.2x3mm"
    SMD6_6x6_4mm = "SMD,6.6x6.4mm"
    SMD1_2x0_6mm = "SMD,1.2x0.6mm"
    SMD1_2x1_8mm = "SMD,1.2x1.8mm"
    SMD5x5_2mm = "SMD,5x5.2mm"
    SMD8_3x8_3mm = "SMD,8.3x8.3mm"
    SMD10_2x10_8mm = "SMD,10.2x10.8mm"
    SMD2_5x3_2mm = "SMD,2.5x3.2mm"
    SMD4x4_5mm = "SMD,4x4.5mm"
    SMD8_5x8mm = "SMD,8.5x8mm"
    SMD3_5x3_2mm = "SMD,3.5x3.2mm"
    SMD12_9x13_2mm = "SMD,12.9x13.2mm"
    SMD8_8x8_2mm = "SMD,8.8x8.2mm"
    SMD4_1x4_4mm = "SMD,4.1x4.4mm"
    SMD10_8x10mm = "SMD,10.8x10mm"
    SMD10_5x10mm = "SMD,10.5x10mm"
    SMD1_1x1_8mm = "SMD,1.1x1.8mm"
    SMD7_5x7mm = "SMD,7.5x7mm"
    SMD3_8x3_8mm = "SMD,3.8x3.8mm"
    SMD1_6x2_2mm = "SMD,1.6x2.2mm"
    SMD12_2x12_2mm = "SMD,12.2x12.2mm"
    SMD4_8x4_8mm = "SMD,4.8x4.8mm"
    SMD5_7x5_4mm = "SMD,5.7x5.4mm"
    SMD15_5x16_5mm = "SMD,15.5x16.5mm"
    SMD8_2x8_8mm = "SMD,8.2x8.8mm"
    SMD10x14mm = "SMD,10x14mm"
    SMD5_1x5_4mm = "SMD,5.1x5.4mm"
    SMD16_5x15_5mm = "SMD,16.5x15.5mm"
    SMD2_1x3mm = "SMD,2.1x3mm"
    SMD10x11_6mm = "SMD,10x11.6mm"
    SMD3_2x4mm = "SMD,3.2x4mm"
    SMD7_2x7_9mm = "SMD,7.2x7.9mm"
    SMD5_8x5_8mm = "SMD,5.8x5.8mm"
    SMD6_6x7_4mm = "SMD,6.6x7.4mm"
    SMD12_7x12_7mm = "SMD,12.7x12.7mm"
    SMD1_2x2mm = "SMD,1.2x2mm"
    SMD1x1_7mm = "SMD,1x1.7mm"
    SMD4_4x4_1mm = "SMD,4.4x4.1mm"
    SMD4_2x4_2mm = "SMD,4.2x4.2mm"

    @staticmethod
    def table(
        to_metric: bool,
    ) -> "dict[SMDSize, SMDSize]":
        table = {
            SMDSize.M0402: SMDSize.I01005,
            SMDSize.M0603: SMDSize.I0201,
            SMDSize.M1005: SMDSize.I0402,
            SMDSize.M1608: SMDSize.I0603,
            SMDSize.M2012: SMDSize.I0805,
            SMDSize.M3216: SMDSize.I1206,
            SMDSize.M3225: SMDSize.I1210,
            SMDSize.M4520: SMDSize.I1808,
            SMDSize.M4532: SMDSize.I1812,
            SMDSize.M4564: SMDSize.I1825,
            SMDSize.M5750: SMDSize.I2220,
            SMDSize.M5664: SMDSize.I2225,
            SMDSize.M9110: SMDSize.I3640,
        }

        if to_metric:
            return {v: k for k, v in table.items()}
        return table

    @property
    def imperial(self) -> "SMDSize":
        if self.name.startswith("I"):
            return self

        try:
            return self.table(to_metric=False)[self]
        except KeyError:
            raise self.UnableToConvert(f"Unable to convert {self} to imperial")

    @property
    def metric(self) -> "SMDSize":
        if self.name.startswith("M"):
            return self

        try:
            return self.table(to_metric=True)[self]
        except KeyError:
            raise self.UnableToConvert(f"Unable to convert {self} to metric")

    @property
    def without_prefix(self) -> "str":
        name = self.name
        for prefix in ["I", "M"]:
            name = name.removeprefix(prefix)
        return name



# This file is part of the faebryk project
# SPDX-License-Identifier: MIT

from textwrap import dedent

from faebryk.libs.codegen.atocodegen import AtoCodeGen
from faebryk.libs.codegen.atocodeparse import AtoCodeParse


def _cleanup(s: str) -> str:
    return dedent(s).removeprefix("\n")


def test_atogen_basic_part():
    cf = AtoCodeGen.ComponentFile("TI_TCA9548APWR")

    cf.add_trait(
        "is_atomic_part",
        manufacturer="TI",
        partnumber="TCA9548APWR",
        footprint="TSSOP-24_L7.8-W4.4-P0.65-LS6.4-BL.kicad_mod",
        symbol="TCA9548APWR.kicad_sym",
        model="TSSOP-24_L7.8-W4.4-H1.0-LS6.4-P0.65.step",
    )
    cf.add_comments(
        "This trait marks this file as auto-generated",
        "If you want to manually change it, remove the trait",
        use_spacer=True,
    )
    cf.add_trait(
        "is_auto_generated",
        system="ato_part",
        source="easyeda:C130026",
        date="2025-05-26T12:53:38.352371+00:00",
        checksum="{CHECKSUM_PLACEHOLDER}",
    )
    ato = cf.dump()

    assert ato == _cleanup(
        """
        #pragma experiment("TRAITS")
        import is_atomic_part
        import is_auto_generated

        component TI_TCA9548APWR:
            trait is_atomic_part<manufacturer="TI", partnumber="TCA9548APWR", footprint="TSSOP-24_L7.8-W4.4-P0.65-LS6.4-BL.kicad_mod", symbol="TCA9548APWR.kicad_sym", model="TSSOP-24_L7.8-W4.4-H1.0-LS6.4-P0.65.step">

            # This trait marks this file as auto-generated
            # If you want to manually change it, remove the trait
            trait is_auto_generated<system="ato_part", source="easyeda:C130026", date="2025-05-26T12:53:38.352371+00:00", checksum="{CHECKSUM_PLACEHOLDER}">
        """  # noqa: E501
    )


def test_atogen_postmod():
    cf = AtoCodeGen.ComponentFile("BLA")
    t = cf.add_trait("is_atomic_part")

    assert cf.dump() == _cleanup(
        """
        #pragma experiment("TRAITS")
        import is_atomic_part

        component BLA:
            trait is_atomic_part
        """
    )

    t.args = {
        "manufacturer": "TI",
        "partnumber": "TCA9548APWR",
    }

    assert cf.dump() == _cleanup(
        """
        #pragma experiment("TRAITS")
        import is_atomic_part

        component BLA:
            trait is_atomic_part<manufacturer="TI", partnumber="TCA9548APWR">
        """
    )


def test_parse_regression_1():
    file = _cleanup("""
        #pragma experiment("TRAITS")
        import has_designator_prefix
        import has_part_picked
        import is_atomic_part
        import is_auto_generated

        component TDK_INVENSENSE_ICM_20948_package:
            # This trait marks this file as auto-generated
            # If you want to manually change it, remove the trait
            trait is_auto_generated<system="ato_part", source="easyeda:C726001", date="2025-07-08T23:04:40.873125+00:00", checksum="fa1d904cee1d545d9f702fcbe5be0266192efcbe8edc94acbbaa3493bf3aed62">

            trait is_atomic_part<manufacturer="TDK INVENSENSE", partnumber="ICM-20948", footprint="QFN-24_L3.0-W3.0-P0.40-BL-EP.kicad_mod", symbol="ICM-20948.kicad_sym", model="QFN-24_L3.0-W3.0-H0.9-P0.40-BL-EP.step">
            trait has_part_picked::by_supplier<supplier_id="lcsc", supplier_partno="C726001", manufacturer="TDK INVENSENSE", partno="ICM-20948">
            trait has_designator_prefix<prefix="U">

            # pins
            signal AUX_CL ~ pin 7
            signal AUX_DA ~ pin 21
            signal EP ~ pin 25
            signal FSYNC ~ pin 11
            signal GND ~ pin 18
            signal INT1 ~ pin 12
            signal REGOUT ~ pin 10
            signal RESV ~ pin 19
            RESV ~ pin 20
            signal SCL_SCLK ~ pin 23
            signal SDA_SDI ~ pin 24
            signal SDO_AD0 ~ pin 9
            signal VDD ~ pin 13
            signal VDDIO ~ pin 8
            signal nCS ~ pin 22
        """)  # noqa: E501

    ato = AtoCodeParse.ComponentFile(file)

    assert ato.parse_trait("is_atomic_part") == (
        None,
        {
            "manufacturer": "TDK INVENSENSE",
            "partnumber": "ICM-20948",
            "footprint": "QFN-24_L3.0-W3.0-P0.40-BL-EP.kicad_mod",
            "symbol": "ICM-20948.kicad_sym",
            "model": "QFN-24_L3.0-W3.0-H0.9-P0.40-BL-EP.step",
        },
    )

    import faebryk.library._F as F

    atomic_trait = ato.get_trait(F.is_atomic_part)
    assert atomic_trait._manufacturer == "TDK INVENSENSE"
    assert atomic_trait._partnumber == "ICM-20948"
    assert atomic_trait._footprint == "QFN-24_L3.0-W3.0-P0.40-BL-EP.kicad_mod"
    assert atomic_trait._symbol == "ICM-20948.kicad_sym"
    assert atomic_trait._model == "QFN-24_L3.0-W3.0-H0.9-P0.40-BL-EP.step"



# This file is part of the faebryk project
# SPDX-License-Identifier: MIT

import unittest

import pytest

import faebryk.libs.picker.lcsc as lcsc

"""
This mode is for when you want to check the models in kicad.
This is especially useful while reverse engineering the easyeda translations.
"""
INTERACTIVE_TESTING = False


@pytest.mark.usefixtures("setup_project_config")
class TestLCSC(unittest.TestCase):
    def test_model_translations(self):
        test_parts = {
            # Zero SMD
            "C1525": (0, 0, 0),
            "C2827654": (0, 0, 0),
            "C284656": (0, 0, 0),
            "C668207": (0, 0, 0),
            "C914087": (0, 0, 0),
            "C25076": (0, 0, 0),
            "C328302": (0, 0, 0),
            "C72041": (0, 0, 0),
            "C99652": (0, 0, 0),
            "C72038": (0, 0, 0),
            "C25111": (0, 0, 0),
            "C2290": (0, 0, 0),
            # Non-zero SMD
            # "C585890": (0, -3.5, 0), # TODO enable
            # "C2828092": (0, -9.4, -0.254),  # TODO enable
            # THT
            # "C5239862": (-2.159, 0, 0), #TODO enable
            # "C225521": (-6.3, 1.3, 0),  # TODO enable
        }

        # if not INTERACTIVE_TESTING:
        #    lcsc.BUILD_FOLDER = Path(mkdtemp())

        for partid, expected in test_parts.items():
            part = lcsc.download_easyeda_info(
                partid,
                get_model=INTERACTIVE_TESTING,
            )

            translation = part.footprint.footprint.footprint.models[0].offset.xyz

            self.assertEqual(
                (translation.x, translation.y, translation.z), expected, f"{partid}"
            )



import logging
from pathlib import Path

from pydantic import BaseModel

from atopile import buildutil
from atopile.cli.logging_ import capture_logs, log_exceptions
from atopile.mcp.util import MCPTools

cli_tools = MCPTools()

logger = logging.getLogger(__name__)


class Result(BaseModel):
    success: bool
    project_dir: str


class ErrorResult(Result):
    error: str
    error_message: str


class BuildResult(Result):
    target: str
    logs: str


@cli_tools.register()
def build_project(
    absolute_project_dir: Path, target_name_from_yaml: str
) -> BuildResult:
    from atopile.build import init_app
    from atopile.config import config

    config.apply_options(
        entry=None,
        working_dir=absolute_project_dir,
        selected_builds=[target_name_from_yaml],
    )

    config.project.open_layout_on_build = False
    config.interactive = False

    success = True

    with config.select_build(target_name_from_yaml), capture_logs() as logs:
        logger.info("Building target '%s'", config.build.name)

        try:
            with log_exceptions(logs):
                app = init_app()
                buildutil.build(app)
        except Exception:
            success = False

    return BuildResult(
        success=success,
        project_dir=str(absolute_project_dir),
        target=target_name_from_yaml,
        logs=logs.getvalue(),
    )


class CreatePartResult(Result):
    manufacturer: str
    part_number: str
    description: str
    supplier_id: str
    stock: int
    path: str
    import_statement: str


class CreatePartError(ErrorResult):
    error: str
    error_message: str


@cli_tools.register()
def search_and_install_jlcpcb_part(
    absolute_project_dir: Path, lcsc_part_number: str
) -> CreatePartResult | CreatePartError:
    """
    Search for a part on JLCPCB and install it.
    """

    from atopile.cli.create import part
    from atopile.config import config

    try:
        apart, component = part(
            search_term=lcsc_part_number,
            accept_single=True,
            project_dir=absolute_project_dir,
        )
    except Exception as e:
        return CreatePartError(
            success=False,
            project_dir=str(absolute_project_dir),
            error=e.__class__.__name__,
            error_message=str(e),
        )

    return CreatePartResult(
        success=True,
        project_dir=str(absolute_project_dir),
        manufacturer=component.manufacturer_name,
        part_number=component.part_number,
        description=component.description,
        supplier_id=component.lcsc_display,
        stock=component.stock,
        path=str(apart.path),
        import_statement=apart.generate_import_statement(config.project.paths.src),
    )


class InstallPackageResult(Result):
    installed_packages: list[str]


class InstallPackageError(ErrorResult):
    pass


@cli_tools.register()
def install_package(
    absolute_project_dir: str,
    package_identifiers: list[str],
    allow_upgrade: bool = False,
) -> InstallPackageResult | InstallPackageError:
    """
    Install a package using the ato CLI.
    """

    from atopile.cli.install import add

    try:
        add(
            package=package_identifiers,
            path=Path(absolute_project_dir),
            upgrade=allow_upgrade,
        )
    except Exception as e:
        return InstallPackageError(
            success=False,
            project_dir=str(absolute_project_dir),
            error=e.__class__.__name__,
            error_message=str(e),
        )

    return InstallPackageResult(
        success=True,
        project_dir=str(absolute_project_dir),
        installed_packages=package_identifiers,
    )



import pytest

from faebryk.libs.util import yield_missing


def test_basic():
    existing = {1, 2, 3}
    candidates = [1, 2, 3, 4, 5]
    assert list(yield_missing(existing, candidates)) == [4, 5]


def test_no_candidates():
    existing = {0, 1, 3}
    missing = iter(yield_missing(existing))
    assert next(missing) == 2
    assert next(missing) == 4
    assert next(missing) == 5
    assert next(missing) == 6


def test_max_range():
    existing = {1, 2, 3}
    missing = iter(yield_missing(existing, range(7)))
    assert next(missing) == 0
    assert next(missing) == 4
    assert next(missing) == 5
    assert next(missing) == 6
    with pytest.raises(StopIteration):
        next(missing)



from atopile.mcp.tools.cli import cli_tools
from atopile.mcp.tools.library import library_tools
from atopile.mcp.tools.packages import packages_tools
from atopile.mcp.tools.project import project_tools

__all__ = ["cli_tools", "library_tools", "packages_tools", "project_tools"]



# This file is part of the faebryk project
# SPDX-License-Identifier: MIT
import logging

import faebryk.library._F as F
from faebryk.core.moduleinterface import ModuleInterface
from faebryk.libs.library import L

logger = logging.getLogger(__name__)


class PDM(ModuleInterface):
    """
    Pulse Density Modulation is a way of representing a sampled signal as a stream of
    single bits where the relative density of the pulses correspond to the analog
    signal's amplitude
    """

    data: F.ElectricLogic
    clock: F.ElectricLogic
    select: F.ElectricLogic

    @L.rt_field
    def single_electric_reference(self):
        return F.has_single_electric_reference_defined(
            F.ElectricLogic.connect_all_module_references(self)
        )



# This file is part of the faebryk project
# SPDX-License-Identifier: MIT

import logging
from itertools import chain, pairwise

import pytest

import faebryk.library._F as F
from faebryk.core.link import (
    LinkDirect,
    LinkDirectConditional,
    LinkDirectConditionalFilterResult,
    LinkDirectDerived,
)
from faebryk.core.module import Module
from faebryk.core.moduleinterface import IMPLIED_PATHS, ModuleInterface
from faebryk.core.node import NodeException
from faebryk.libs.app.erc import (
    ERCFaultShortedModuleInterfaces,
    ERCPowerSourcesShortedError,
    simple_erc,
)
from faebryk.libs.library import L
from faebryk.libs.util import cast_assert, times
from test.common.resources.fabll_modules.ButtonCell import ButtonCell
from test.common.resources.fabll_modules.RP2040 import RP2040
from test.common.resources.fabll_modules.RP2040_ReferenceDesign import (
    RP2040_ReferenceDesign,
)

logger = logging.getLogger(__name__)


def test_self():
    mif = ModuleInterface()
    assert mif.is_connected_to(mif)


def test_up_connect_simple_single():
    """
    ```
    H1      H2
     L1 -->  L1
    ```
    """

    class High(ModuleInterface):
        lower: ModuleInterface

    high1 = High()
    high2 = High()

    high1.lower.connect(high2.lower)
    assert high1.is_connected_to(high2)


def test_up_connect_simple_two():
    """
    ```
    H1      H2
     L1 -->  L1
     L2 -->  L2
    ```
    """

    class High(ModuleInterface):
        lower1: ModuleInterface
        lower2: ModuleInterface

    high1 = High()
    high2 = High()

    high1.lower1.connect(high2.lower1)
    high1.lower2.connect(high2.lower2)
    assert high1.is_connected_to(high2)


def test_up_connect_simple_multiple():
    """
    ```
    H1      H2
     L1 -->  L1
     L2 -->  L2
     L3 -->  L3
    ```
    """

    class High(ModuleInterface):
        lower1: ModuleInterface
        lower2: ModuleInterface
        lower3: ModuleInterface

    high1 = High()
    high2 = High()

    high1.lower1.connect(high2.lower1)
    high1.lower2.connect(high2.lower2)
    high1.lower3.connect(high2.lower3)
    assert high1.is_connected_to(high2)


def test_up_connect_chain_simple():
    """
    ```
    H1            H2
     L1 --> M -->  L1
     L2 -------->  L2
    ```
    """

    class High(ModuleInterface):
        lower1: ModuleInterface
        lower2: ModuleInterface

    high1 = High()
    high2 = High()

    middle = ModuleInterface()

    high1.lower1.connect(middle)
    high2.lower1.connect(middle)
    high1.lower2.connect(high2.lower2)

    assert high1.is_connected_to(high2)


def test_up_connect_chain_multiple_same():
    """
    ```
    H1      H2      H3
     L1 -->  L1 -->  L1
     L2 -->  L2 -->  L2
    ```
    """

    class High(ModuleInterface):
        lower1: ModuleInterface
        lower2: ModuleInterface

    high1 = High()
    high2 = High()
    high3 = High()

    high1.lower1.connect(high2.lower1)
    high1.lower2.connect(high2.lower2)
    high2.lower1.connect(high3.lower1)
    high2.lower2.connect(high3.lower2)

    assert high1.is_connected_to(high3)


def test_up_connect_chain_multiple_mixed():
    """
    ```
    H1      H2  ==>  H3      H4
     L1 -->  L1       L1 -->  L1
     L2 -->  L2       L2 -->  L2
    ```
    """

    class Low(ModuleInterface): ...

    class High(ModuleInterface):
        lower1: Low
        lower2: Low

    high1, high2, high3, high4 = times(4, High)

    high1.lower1.connect(high2.lower1)
    high1.lower2.connect(high2.lower2)
    high2.connect_shallow(high3)
    high3.lower1.connect(high4.lower1)
    high3.lower2.connect(high4.lower2)

    assert high1.is_connected_to(high4)


def test_split_chain_single():
    """
    Miro: Implied bus connection 2
    ```
    H1     H2 --> H3
     L1 --> L1     L1
     L2     L2     L2
      |             ^
      +-------------+
    ```
    """

    class Low(ModuleInterface): ...

    class High(ModuleInterface):
        lower1: Low
        lower2: Low

    high1, high2, high3 = times(3, High)

    high1.lower1.connect(high2.lower1)
    high1.lower2.connect(high3.lower2)
    high2.connect(high3)

    assert high1.is_connected_to(high3)


@pytest.mark.xfail(reason="No support atm for split chains with ambiguous split/hier")
def test_split_chain_double_flat_no_inter():
    """
    ```
    H1     H2 --> H3     H4
     L1 --> L1     L1 --> L1
     L2     L2     L2     L2
      |                    ^
      +--------------------+
    ```
    """

    class Low(ModuleInterface): ...

    class High(ModuleInterface):
        lower1: Low
        lower2: Low

    class App(Module):
        high = L.list_field(4, High)

    app = App()

    high1, high2, high3, high4 = app.high

    high1.lower1.connect(high2.lower1)
    high1.lower2.connect(high4.lower2)
    high2.connect(high3)
    high3.lower1.connect(high4.lower1)

    assert high1.lower1.is_connected_to(high4.lower1)
    x = high1.is_connected_to(high3)
    print(x)
    assert not x
    # assert not high1.is_connected_to(high3)
    assert high1.is_connected_to(high4)

    # TODO: See pathfinder.cpp:67 for failure


def test_split_chain_double_flat_inter():
    # TODO this test is not difficult enough
    # the intermediate is trivially connected since the double split is resolved
    # maybe insert extra node between H2 and H3?
    """
    ```
    H1     H2 --> H3     H4
     L1 --> L1     L1 --> L1
     L2     L2     L2 --> L2
      |                    ^
      +--------------------+
    ```
    """

    class Low(ModuleInterface): ...

    class High(ModuleInterface):
        lower1: Low
        lower2: Low

    class App(Module):
        high = L.list_field(4, High)

    app = App()

    high1, high2, high3, high4 = app.high

    high1.lower1.connect(high2.lower1)
    high1.lower2.connect(high4.lower2)
    high2.connect(high3)
    high3.lower1.connect(high4.lower1)
    high3.lower2.connect(high4.lower2)

    assert high1.lower1.is_connected_to(high4.lower1)
    assert high1.is_connected_to(high3)
    assert high1.is_connected_to(high4)


@pytest.mark.xfail(reason="No support atm for split chains with ambiguous split/hier")
def test_split_chain_double_hierarchy():
    """
    ```
                 R1 --> R2
    H1     H2 --> H      H      H3
     L1 --> L1     L1     L1 --> L1
     L2     L2     L2     L2     L2
      |                           ^
      +---------------------------+
    ```
    """

    class Low(ModuleInterface): ...

    class High(ModuleInterface):
        lower1: Low
        lower2: Low

    class Higher(ModuleInterface):
        high: High

    class App(Module):
        high = L.list_field(3, High)
        higher = L.list_field(2, Higher)

    app = App()

    high1, high2, high3 = app.high
    higher1, higher2 = app.higher

    high1.lower1.connect(higher1.high.lower1)
    high1.lower2.connect(high3.lower2)
    high2.connect(higher1.high)
    higher1.connect_shallow(higher2)
    higher2.high.lower1.connect(high3.lower1)

    assert high1.lower1.is_connected_to(high3.lower1)
    assert high1.is_connected_to(high3)


def test_split_chain_flip():
    """
    Miro: Implied Double-Flip Bus Connection
    ```
    H1     H2 ==> H3     H4
     L1 --> L2     L2 --> L1
     L2 --> L1     L1 --> L2
    ```
    Note: Shallowness not important, just makes it harder
    """

    class Low(ModuleInterface): ...

    class High(ModuleInterface):
        lower1: Low
        lower2: Low

    class App(Module):
        high = L.list_field(4, High)

    app = App()

    high1, high2, high3, high4 = app.high

    high1.lower1.connect(high2.lower2)
    high1.lower2.connect(high2.lower1)
    high2.connect_shallow(high3)
    high3.lower1.connect(high4.lower2)
    high3.lower2.connect(high4.lower1)

    assert high1.is_connected_to(high4)


def test_split_flip_negative():
    """
    Miro: Implied Bus Non-Connection
    ```
    H1     H2
     L1 --> L2
     L2 --> L1
    ```
    """

    class Low(ModuleInterface): ...

    class High(ModuleInterface):
        lower1: Low
        lower2: Low

    high1, high2 = times(2, High)

    high1.lower1.connect(high2.lower2)
    high1.lower2.connect(high2.lower1)

    assert not high1.is_connected_to(high2)


def test_up_connect_chain_multiple_mixed_simulate_realworld():
    """
    ```
    H1      H2  ==>  H3      H4
     L1 -->  L1       L1 -->  L1
     L2 ---  L2 ----- L2 ---  L2
    ```
    """

    class Low(ModuleInterface): ...

    class High(ModuleInterface):
        lower1: Low
        lower2: Low

    high1, high2, high3, high4 = times(4, High)

    high1.lower1.connect(high2.lower1)
    high2.connect_shallow(high3)
    high3.lower1.connect(high4.lower1)

    high1.lower2.connect(high2.lower2, high3.lower2, high4.lower2)

    assert high1.is_connected_to(high4)


@pytest.mark.xfail(reason="No support atm for split chains with ambiguous split/hier")
def test_up_connect_chain_multiple_realworld():
    """
    ```
    L1      L2 ==>  L3     L4
     S -->  S       S -->  S
     R ---  R ----- R ---  R
      HV     HV      HV     HV
      LV     LV      LV     LV
    ```
    """

    l1, l2, l3, l4 = times(4, F.ElectricLogic)

    l1.signal.connect(l2.signal)
    l2.connect_shallow(l3)
    l3.signal.connect(l4.signal)

    l1.reference.connect(l2.reference, l3.reference, l4.reference)

    assert l1.is_connected_to(l4)


def test_up_connect_chain_hierarchy():
    """
    ```
    R1              R2
     H1     HM1 ==>  H1
      L1 -->  L1      L1
      L2 -->  L2      L2
     H2 ==> HM2      H2
      L1      L1 -->  L1
      L2      L2 -->  L2
    ```
    """

    class Low(ModuleInterface): ...

    class High(ModuleInterface):
        lower1: Low
        lower2: Low

    class Higher(ModuleInterface):
        high1: High
        high2: High

    higher_begin = Higher()
    higher_end = Higher()

    high_middle1 = High()
    high_middle2 = High()

    higher_begin.high1.lower1.connect(high_middle1.lower1)
    higher_begin.high1.lower2.connect(high_middle1.lower2)
    higher_begin.high2.connect_shallow(high_middle2)
    higher_end.high1.connect_shallow(high_middle1)
    higher_end.high2.lower1.connect(high_middle2.lower1)
    higher_end.high2.lower2.connect(high_middle2.lower2)

    assert higher_begin.is_connected_to(higher_end)


def test_up_connect_hierarchy():
    """
    ```
    R1      R2
     H1      H1
      L1 -->  L1
      L2 -->  L2
     H2      H2
      L1 -->  L1
      L2 -->  L2
    ```
    """

    class High(ModuleInterface):
        lower1: ModuleInterface
        lower2: ModuleInterface

    class Higher(ModuleInterface):
        high1: High
        high2: High

    higher1 = Higher()
    higher2 = Higher()

    higher1.high1.lower1.connect(higher2.high1.lower1)
    higher1.high1.lower2.connect(higher2.high1.lower2)
    higher1.high2.lower1.connect(higher2.high2.lower1)
    higher1.high2.lower2.connect(higher2.high2.lower2)
    assert higher1.is_connected_to(higher2)


def test_up_connect_hierarchy_mixed():
    """
    ```
    R1      R2
     H1      H1
      L1 -->  L1
      L2 -->  L2
     H2 ==>  H2
      L1      L1
      L2      L2
    ```
    """

    class Low(ModuleInterface): ...

    class High(ModuleInterface):
        lower1: Low
        lower2: Low

    class Higher(ModuleInterface):
        high1: High
        high2: High

    higher1 = Higher()
    higher2 = Higher()

    higher1.high1.lower1.connect(higher2.high1.lower1)
    higher1.high1.lower2.connect(higher2.high1.lower2)
    higher1.high2.connect_shallow(higher2.high2)
    assert higher1.is_connected_to(higher2)


def test_up_connect_simple_two_negative():
    """
    ```
    H1      H2
     L1 -->  L1
     L2      L2
    ```
    """

    class High(ModuleInterface):
        lower1: ModuleInterface
        lower2: ModuleInterface

    high1 = High()
    high2 = High()

    high1.lower1.connect(high2.lower1)
    assert not high1.is_connected_to(high2)


def test_up_connect_simple_multiple_negative():
    """
    ```
    H1      H2
     L1 -->  L1
     L2 -->  L2
     L3      L3
    ```
    """

    class High(ModuleInterface):
        lower1: ModuleInterface
        lower2: ModuleInterface
        lower3: ModuleInterface

    high1 = High()
    high2 = High()

    high1.lower1.connect(high2.lower1)
    high1.lower2.connect(high2.lower2)
    assert not high1.is_connected_to(high2)


def test_up_connect():
    """
    ```
    BI     BO
     RX      RX
      S -->  S
      R -->  R
       HV      HV
       LV      LV
     TX      TX
      S -->  S
      R -->  R
       HV      HV
       LV      LV
    ```
    """

    class UARTBuffer(Module):
        bus_in: F.UART_Base
        bus_out: F.UART_Base

        def __preinit__(self) -> None:
            self.bus_in.rx.line.connect(self.bus_out.rx.line)
            self.bus_in.tx.line.connect(self.bus_out.tx.line)
            self.bus_in.rx.reference.connect(self.bus_out.rx.reference)

    app = UARTBuffer()

    assert app.bus_in.rx.line.is_connected_to(app.bus_out.rx.line)
    assert app.bus_in.rx.reference.is_connected_to(app.bus_out.rx.reference)
    assert app.bus_in.rx.is_connected_to(app.bus_out.rx)
    assert app.bus_in.tx.is_connected_to(app.bus_out.tx)
    assert app.bus_in.is_connected_to(app.bus_out)


def test_down_connect():
    """
    ```
    P1 -->  P2
     HV      HV
     LV      LV
    ```
    """

    ep = times(2, F.ElectricPower)
    ep[0].connect(ep[1])

    assert ep[0].is_connected_to(ep[1])
    assert ep[0].hv.is_connected_to(ep[1].hv)
    assert ep[0].lv.is_connected_to(ep[1].lv)


def test_chains_direct():
    """
    ```
    M1 --> M2 --> M3
    ```
    """

    mifs = times(3, ModuleInterface)
    mifs[0].connect(mifs[1])
    mifs[1].connect(mifs[2])
    assert mifs[0].is_connected_to(mifs[2])


def test_chains_double_shallow_flat():
    """
    ```
    M1 ==> M2 ==> M3
    ```
    """

    mifs = times(3, ModuleInterface)
    mifs[0].connect_shallow(mifs[1])
    mifs[1].connect_shallow(mifs[2])
    assert mifs[0].is_connected_to(mifs[2])


def test_chains_mixed_shallow_flat():
    """
    ```
    M1 ==> M2 --> M3
    ```
    """

    mifs = times(3, ModuleInterface)
    mifs[0].connect_shallow(mifs[1])
    mifs[1].connect(mifs[2])
    assert mifs[0].is_connected_to(mifs[2])


def test_chains_mixed_shallow_nested():
    """
    ```
    L1  ==>  L2 -->  L3
     S        S       S
     R        R       R
      HV       HV      HV
      LV       LV      LV
    ```
    """
    # Test hierarchy down filter & chain resolution
    el = times(3, F.ElectricLogic)
    el[0].connect_shallow(el[1])
    el[1].connect(el[2])
    assert el[0].is_connected_to(el[2])

    assert el[1].line.is_connected_to(el[2].line)
    assert el[1].reference.is_connected_to(el[2].reference)
    assert not el[0].line.is_connected_to(el[1].line)
    assert not el[0].reference.is_connected_to(el[1].reference)
    assert not el[0].line.is_connected_to(el[2].line)
    assert not el[0].reference.is_connected_to(el[2].reference)

    # Test duplicate resolution
    el[0].line.connect(el[1].line)
    el[0].reference.connect(el[1].reference)
    assert el[0].is_connected_to(el[1])
    assert el[0].is_connected_to(el[2])


def test_loooooong_chain():
    """Let's make it hard"""
    mifs = times(2**10, F.ElectricPower)
    for left, right in pairwise(mifs):
        left.connect(right)

    assert mifs[0].is_connected_to(mifs[-1])


# FIXME: this should be WAYYY higher than 16
@pytest.mark.parametrize("length", [16])
def test_alternating_long_chain(length):
    """Let's make it hard"""
    mifs = times(length, F.ElectricPower)
    for i, (left, right) in enumerate(pairwise(mifs)):
        if i % 2:
            left.connect(right)
        else:
            left.lv.connect(right.lv)
            left.hv.connect(right.hv)

    assert mifs[0].is_connected_to(mifs[-1])
    assert mifs[0].lv.is_connected_to(mifs[-1].lv)
    assert mifs[0].hv.is_connected_to(mifs[-1].hv)


def test_shallow_bridge_simple():
    """
    ```
                B
    H1      HI ===> HO     H2
     L1 -->  L1      L1 --> L1
     L2 -->  L2      L2 --> L2
    ```
    """

    class Low(ModuleInterface): ...

    class High(ModuleInterface):
        lower1: Low
        lower2: Low

    class ShallowBridge(Module):
        high_in: High
        high_out: High

        def __preinit__(self) -> None:
            self.high_in.connect_shallow(self.high_out)

        @L.rt_field
        def can_bridge(self):
            return F.can_bridge_defined(self.high_in, self.high_out)

    bridge = ShallowBridge()
    high1 = High()
    high2 = High()
    high1.connect_via(bridge, high2)

    assert high1.is_connected_to(high2)
    assert not bridge.high_in.lower1.is_connected_to(bridge.high_out.lower1)
    assert not bridge.high_in.lower2.is_connected_to(bridge.high_out.lower2)
    assert not high1.lower1.is_connected_to(high2.lower1)
    assert not high1.lower2.is_connected_to(high2.lower2)


@pytest.mark.xfail(reason="No support atm for split chains with ambiguous split/hier")
def test_shallow_bridge_partial():
    """
    ```
             ________B__________
     L1          LI ===> LO          L2
      S -->  I -> S       S -> O -->  S
      R --------  R ----- R --------  R
    ```
    """

    class Buffer(Module):
        ins: F.Electrical
        outs: F.Electrical

        ins_l: F.ElectricLogic
        outs_l: F.ElectricLogic

        def __preinit__(self) -> None:
            self.ins_l.signal.connect(self.ins)
            self.outs_l.signal.connect(self.outs)

            self.ins_l.connect_shallow(self.outs_l)

        @L.rt_field
        def single_electric_reference(self):
            return F.has_single_electric_reference_defined(
                F.ElectricLogic.connect_all_module_references(self)
            )

    l1 = F.ElectricLogic()
    l2 = F.ElectricLogic()
    b = Buffer()

    l1.signal.connect(b.ins)
    l2.signal.connect(b.outs)
    l1.reference.connect(b.single_electric_reference.get_reference())
    l2.reference.connect(b.single_electric_reference.get_reference())

    assert l1.is_connected_to(l2)


def test_shallow_bridge_full():
    """
    Test the bridge connection between two UART interfaces through a buffer:

    ```
    U1 ---> _________B________ ---> U2
     TX          LI ===> LO          TX
      S -->  I -> S       S -> O -->  S
      R --------  R ----- R --------  R
    ```

    Where:
    - U1, U2: UART interfaces
    - B: Buffer
    - TX: Transmit
    - S: Signal
    - R: Reference
    - I: Input
    - O: Output
    - IL: Input Logic
    - OL: Output Logic
    """

    class Buffer(Module):
        ins = L.list_field(2, F.Electrical)
        outs = L.list_field(2, F.Electrical)

        ins_l = L.list_field(2, F.ElectricLogic)
        outs_l = L.list_field(2, F.ElectricLogic)

        def __preinit__(self) -> None:
            assert (
                self.ins_l[0].reference
                is self.ins_l[0].single_electric_reference.get_reference()
            )

            for el, lo in chain(
                zip(self.ins, self.ins_l),
                zip(self.outs, self.outs_l),
            ):
                lo.line.connect(el)

            for l1, l2 in zip(self.ins_l, self.outs_l):
                l1.connect_shallow(l2)

        @L.rt_field
        def single_electric_reference(self):
            return F.has_single_electric_reference_defined(
                F.ElectricLogic.connect_all_module_references(self)
            )

    class UARTBuffer(Module):
        buf: Buffer
        bus_in: F.UART_Base
        bus_out: F.UART_Base

        def __preinit__(self) -> None:
            bus_i = self.bus_in
            bus_o = self.bus_out
            buf = self.buf

            bus_i.tx.line.connect(buf.ins[0])
            bus_i.rx.line.connect(buf.ins[1])
            bus_o.tx.line.connect(buf.outs[0])
            bus_o.rx.line.connect(buf.outs[1])

        @L.rt_field
        def single_electric_reference(self):
            return F.has_single_electric_reference_defined(
                F.ElectricLogic.connect_all_module_references(self)
            )

    app = UARTBuffer()

    bus_i = app.bus_in
    bus_o = app.bus_out
    buf = app.buf

    # Check that the two buffer sides are not connected electrically
    assert not buf.ins[0].is_connected_to(buf.outs[0])
    assert not buf.ins[1].is_connected_to(buf.outs[1])
    assert not bus_i.rx.line.is_connected_to(bus_o.rx.line)
    assert not bus_i.tx.line.is_connected_to(bus_o.tx.line)

    # direct connect
    assert bus_i.tx.line.is_connected_to(buf.ins[0])
    assert bus_i.rx.line.is_connected_to(buf.ins[1])
    assert bus_o.tx.line.is_connected_to(buf.outs[0])
    assert bus_o.rx.line.is_connected_to(buf.outs[1])

    # connect through trait
    assert (
        buf.ins_l[0].single_electric_reference.get_reference() is buf.ins_l[0].reference
    )
    assert buf.ins_l[0].reference.is_connected_to(buf.outs_l[0].reference)
    assert buf.outs_l[1].reference.is_connected_to(buf.ins_l[0].reference)
    assert bus_i.rx.reference.is_connected_to(bus_o.rx.reference)

    # connect through up
    assert bus_i.tx.is_connected_to(buf.ins_l[0])
    assert bus_o.tx.is_connected_to(buf.outs_l[0])

    # connect shallow
    assert buf.ins_l[0].is_connected_to(buf.outs_l[0])

    # Check that the two buffer sides are connected logically
    assert bus_i.tx.is_connected_to(bus_o.tx)
    assert bus_i.rx.is_connected_to(bus_o.rx)
    assert bus_i.is_connected_to(bus_o)


class Specialized(ModuleInterface): ...


class DoubleSpecialized(Specialized): ...


def test_specialize_general_to_special():
    # general connection -> specialized connection
    mifs = times(3, ModuleInterface)
    mifs_special = times(3, Specialized)

    mifs[0].connect(mifs[1])
    mifs[1].connect(mifs[2])

    mifs[0].specialize(mifs_special[0])
    mifs[2].specialize(mifs_special[2])

    assert mifs_special[0].is_connected_to(mifs_special[2])


def test_specialize_special_to_general():
    # specialized connection -> general connection
    mifs = times(3, ModuleInterface)
    mifs_special = times(3, Specialized)

    mifs_special[0].connect(mifs_special[1])
    mifs_special[1].connect(mifs_special[2])

    mifs[0].specialize(mifs_special[0])
    mifs[2].specialize(mifs_special[2])

    assert mifs[0].is_connected_to(mifs[2])


def test_specialize_link():
    # test special link
    class _Link(LinkDirectConditional):
        def __init__(self):
            super().__init__(
                lambda path: LinkDirectConditionalFilterResult.FILTER_PASS,
                needs_only_first_in_path=True,
            )

    mifs = times(3, ModuleInterface)
    mifs_special = times(3, Specialized)

    mifs[0].connect(mifs[1], link=_Link)
    mifs[1].connect(mifs[2])

    mifs[0].specialize(mifs_special[0])
    mifs[2].specialize(mifs_special[2])

    assert mifs_special[0].is_connected_to(mifs_special[2])


def test_specialize_double_with_gap():
    # double specialization with gap
    mifs = times(2, ModuleInterface)
    mifs_special = times(1, Specialized)
    mifs_double_special = times(2, DoubleSpecialized)

    mifs[0].connect(mifs[1])
    mifs[0].specialize(mifs_special[0])
    mifs_special[0].specialize(mifs_double_special[0])
    mifs[1].specialize(mifs_double_special[1])

    assert mifs_double_special[0].is_connected_to(mifs_double_special[1])


def test_specialize_double_with_gap_2():
    mifs = times(2, ModuleInterface)
    mifs_special = times(1, Specialized)
    mifs_double_special = times(2, DoubleSpecialized)

    mifs_double_special[0].connect(mifs_double_special[1])
    mifs[0].specialize(mifs_special[0])
    mifs_special[0].specialize(mifs_double_special[0])
    mifs[1].specialize(mifs_double_special[1])

    assert mifs[0].is_connected_to(mifs[1])


def test_specialize_module():
    battery = F.Battery()
    power = F.ElectricPower()

    battery.power.connect(power)
    buttoncell = battery.specialize(ButtonCell())

    assert buttoncell.power.is_connected_to(battery.power)
    assert power.is_connected_to(buttoncell.power)


def test_isolated_connect_simple():
    x1 = F.ElectricLogic()
    x2 = F.ElectricLogic()
    x1.connect(x2, link=F.ElectricLogic.LinkIsolatedReference)

    assert x1.is_connected_to(x2)
    assert x1.line.is_connected_to(x2.line)

    assert not x1.reference.is_connected_to(x2.reference)
    assert not x1.reference.hv.is_connected_to(x2.reference.hv)


def test_isolated_connect_erc():
    y1 = F.ElectricPower()
    y2 = F.ElectricPower()

    y1.make_source()
    y2.make_source()

    with pytest.raises(ERCPowerSourcesShortedError):
        y1.connect(y2)
        simple_erc(y1.get_graph())

    ldo1 = F.LDO()
    ldo2 = F.LDO()

    with pytest.raises(ERCPowerSourcesShortedError):
        ldo1.power_out.connect(ldo2.power_out)
        simple_erc(ldo1.get_graph())

    a1 = F.I2C()
    b1 = F.I2C()

    a1.connect(b1, link=F.ElectricLogic.LinkIsolatedReference)
    assert a1.is_connected_to(b1)
    assert a1.scl.line.is_connected_to(b1.scl.line)
    assert a1.sda.line.is_connected_to(b1.sda.line)

    assert not a1.scl.reference.is_connected_to(b1.scl.reference)
    assert not a1.sda.reference.is_connected_to(b1.sda.reference)


def test_simple_erc_ElectricPower_short():
    ep1 = F.ElectricPower()
    ep2 = F.ElectricPower()

    ep1.connect(ep2)

    # This is okay!
    simple_erc(ep1.get_graph())

    ep1.lv.connect(ep2.hv)

    # This is not okay!
    with pytest.raises(ERCFaultShortedModuleInterfaces) as ex:
        simple_erc(ep1.get_graph())

    assert set(ex.value.path) == {ep1.lv, ep2.hv}


@pytest.mark.skipif(not IMPLIED_PATHS, reason="IMPLIED_PATHS is not set")
def test_direct_implied_paths():
    powers = times(2, F.ElectricPower)

    # direct implied
    powers[0].connect(powers[1])

    assert powers[1].hv in powers[0].hv.get_connected()

    paths = powers[0].hv.is_connected_to(powers[1].hv)
    assert paths
    path = paths[0]
    assert len(path) == 4
    assert isinstance(path[1].is_connected_to(path[2]), LinkDirectDerived)


@pytest.mark.skipif(not IMPLIED_PATHS, reason="IMPLIED_PATHS is not set")
def test_children_implied_paths():
    powers = times(3, F.ElectricPower)

    # children implied
    powers[0].connect(powers[1])
    powers[1].hv.connect(powers[2].hv)
    powers[1].lv.connect(powers[2].lv)

    assert powers[2] in powers[0].get_connected()

    paths = list(powers[0].is_connected_to(powers[2]))
    assert paths
    assert len(paths[0]) == 4
    assert isinstance(paths[0][1].is_connected_to(paths[0][2]), LinkDirectDerived)


@pytest.mark.skipif(not IMPLIED_PATHS, reason="IMPLIED_PATHS is not set")
def test_shallow_implied_paths():
    powers = times(4, F.ElectricPower)

    # shallow implied
    powers[0].connect(powers[1])
    powers[1].hv.connect(powers[2].hv)
    powers[1].lv.connect(powers[2].lv)
    powers[2].connect_shallow(powers[3])

    assert powers[3] in powers[0].get_connected()

    assert not powers[0].hv.is_connected_to(powers[3].hv)


def test_direct_shallow_instance():
    class MIFType(ModuleInterface):
        pass

    mif1 = MIFType()
    mif2 = MIFType()
    mif3 = MIFType()

    mif1.connect_shallow(mif2, mif3)
    assert isinstance(
        mif1.connected.is_connected_to(mif2.connected), MIFType.LinkDirectShallow()
    )
    assert isinstance(
        mif1.connected.is_connected_to(mif3.connected), MIFType.LinkDirectShallow()
    )


def test_regression_rp2040_usb_diffpair_minimal():
    usb = F.USB2_0_IF.Data()
    terminated_usb = usb.terminated()

    other_usb = F.USB2_0_IF.Data()
    terminated_usb.connect(other_usb)

    n_ref = usb.n.reference
    p_ref = usb.p.reference
    t_n_ref = terminated_usb.n.reference
    t_p_ref = terminated_usb.p.reference
    o_n_ref = other_usb.n.reference
    o_p_ref = other_usb.p.reference
    refs = {n_ref, p_ref, t_n_ref, t_p_ref, o_n_ref, o_p_ref}

    assert isinstance(
        usb.connected.is_connected_to(terminated_usb.connected),
        F.USB2_0_IF.Data.LinkDirectShallow(),
    )
    assert isinstance(
        other_usb.connected.is_connected_to(terminated_usb.connected), LinkDirect
    )
    assert usb.connected.is_connected_to(other_usb.connected) is None

    connected_per_mif = {ref: ref.get_connected(include_self=True) for ref in refs}

    assert not {n_ref, p_ref} & connected_per_mif[t_n_ref].keys()
    assert not {n_ref, p_ref} & connected_per_mif[t_p_ref].keys()
    assert not {t_n_ref, t_p_ref} & connected_per_mif[n_ref].keys()
    assert not {t_n_ref, t_p_ref} & connected_per_mif[p_ref].keys()

    assert set(connected_per_mif[n_ref].keys()) == {n_ref, p_ref}
    assert set(connected_per_mif[p_ref].keys()) == {n_ref, p_ref}
    assert set(connected_per_mif[t_n_ref].keys()) == {
        t_n_ref,
        t_p_ref,
        o_n_ref,
        o_p_ref,
    }
    assert set(connected_per_mif[t_p_ref].keys()) == {
        t_n_ref,
        t_p_ref,
        o_n_ref,
        o_p_ref,
    }

    # close references
    p_ref.connect(other_usb.p.reference)

    connected_per_mif_post = {ref: ref.get_connected(include_self=True) for ref in refs}
    for _, connected in connected_per_mif_post.items():
        assert set(connected.keys()).issuperset(refs)


def test_regression_rp2040_usb_diffpair():
    app = RP2040_ReferenceDesign()

    terminated_usb = cast_assert(F.USB2_0_IF.Data, app.runtime["_terminated_usb_data"])
    rp_usb = app.rp2040.usb

    t_p_ref = terminated_usb.p.reference
    t_n_ref = terminated_usb.n.reference
    r_p_ref = rp_usb.p.reference
    r_n_ref = rp_usb.n.reference
    refs = [
        r_p_ref,
        r_n_ref,
        t_p_ref,
        t_n_ref,
    ]

    connected_per_mif = {ref: ref.get_connected(include_self=True) for ref in refs}
    for connected in connected_per_mif.values():
        assert set(connected.keys()) == set(refs)


@pytest.mark.slow
def test_regression_rp2040_usb_diffpair_full():
    app = RP2040_ReferenceDesign()
    rp2040_2 = RP2040()
    rp2040_3 = RP2040()

    # make graph bigger
    app.rp2040.i2c[0].connect(rp2040_2.i2c[0])
    app.rp2040.i2c[0].connect(rp2040_3.i2c[0])

    F.is_bus_parameter.resolve_bus_parameters(app.get_graph())


def test_connect_incompatible():
    class A(ModuleInterface):
        pass

    class B(ModuleInterface):
        pass

    x = A()
    y = B()
    with pytest.raises(NodeException):
        x.connect(y)  # type: ignore


def test_connect_incompatible_hierarchical():
    class A(ModuleInterface):
        pass

    class B(A):
        pass

    x = A()
    y = B()
    with pytest.raises(NodeException):
        x.connect(y)  # type: ignore


def test_connect_incompatible_hierarchical_regression():
    x = F.ElectricPower()
    y = F.Electrical()

    with pytest.raises(NodeException):
        x.connect(y)  # type: ignore



# This file is part of the faebryk project
# SPDX-License-Identifier: MIT

import logging
import sys
from pathlib import Path
from typing import TYPE_CHECKING

import pytest

import faebryk.library._F as F
from faebryk.core.module import Module
from faebryk.core.solver.defaultsolver import DefaultSolver
from faebryk.core.solver.nullsolver import NullSolver
from faebryk.libs.library import L
from faebryk.libs.picker.api.picker_lib import (
    NotCompatibleException,
    check_and_attach_candidates,
    get_candidates,
)
from faebryk.libs.picker.lcsc import PickedPartLCSC
from faebryk.libs.picker.picker import PickError, pick_part_recursively
from faebryk.libs.sets.sets import EnumSet
from faebryk.libs.smd import SMDSize
from faebryk.libs.units import P
from faebryk.libs.util import cast_assert, groupby

sys.path.append(str(Path(__file__).parent))

if TYPE_CHECKING:
    from components import ComponentTestCase

try:
    from components import components_to_test
except ImportError:
    components_to_test = []

logger = logging.getLogger(__name__)


def test_load_components():
    assert components_to_test, "Failed to load components"


def _make_id(m: "ComponentTestCase"):
    if m.override_test_name:
        module_name = m.override_test_name
    else:
        module_name = type(m.module).__name__
        gouped_by_type = groupby(components_to_test, lambda c: type(c.module))
        group_for_module = gouped_by_type[type(m.module)]
        if len(group_for_module) > 1:
            module_name += f"[{group_for_module.index(m)}]"

    return module_name


@pytest.mark.usefixtures("setup_project_config")
@pytest.mark.skipif(components_to_test is None, reason="Failed to load components")
@pytest.mark.parametrize(
    "case",
    components_to_test,
    ids=[_make_id(m) for m in components_to_test],
)
def test_pick_module(case: "ComponentTestCase"):
    module = case.module

    if case.packages:
        module.add(F.has_package_requirements(size=EnumSet(*case.packages)))

    # pick
    solver = DefaultSolver()
    pick_part_recursively(module, solver)

    assert module.has_trait(F.has_part_picked)
    part = module.get_trait(F.has_part_picked).get_part()

    # Sanity check
    assert part.partno

    # Check LCSC & MFR
    if case.lcsc_id:
        assert cast_assert(PickedPartLCSC, part).lcsc_id == case.lcsc_id
    elif case.mfr_mpn:
        assert part.manufacturer == case.mfr_mpn[0]
        assert part.partno == case.mfr_mpn[1]

    # Check parameters
    # params = module.get_children(types=Parameter, direct_only=True)
    # TODO check that part params are equal (alias_is) to module params


@pytest.mark.usefixtures("setup_project_config")
def test_type_pick():
    module = F.Resistor()

    assert module.has_trait(F.is_pickable_by_type)
    assert module.has_trait(F.is_pickable)
    module.resistance.constrain_subset(L.Range.from_center_rel(100 * P.ohm, 0.1))

    pick_part_recursively(module, DefaultSolver())

    assert module.has_trait(F.has_part_picked)


@pytest.mark.usefixtures("setup_project_config")
def test_no_pick():
    module = Module()
    module.add(F.has_part_removed())

    pick_part_recursively(module, DefaultSolver())

    assert module.has_trait(F.has_part_picked)
    assert module.get_trait(F.has_part_picked).removed


@pytest.mark.usefixtures("setup_project_config")
def test_no_pick_inherit_override_none():
    class _CapInherit(F.Capacitor):
        pickable = None

    module = _CapInherit()

    assert not module.has_trait(F.is_pickable)

    pick_part_recursively(module, DefaultSolver())

    assert not module.has_trait(F.has_part_picked)


@pytest.mark.usefixtures("setup_project_config")
def test_no_pick_inherit_remove():
    class _(F.Capacitor):
        no_pick: F.has_part_removed

    module = _()

    pick_part_recursively(module, DefaultSolver())

    assert module.has_trait(F.has_part_picked)
    assert module.get_trait(F.has_part_picked).removed


@pytest.mark.usefixtures("setup_project_config")
def test_skip_self_pick():
    # TODO: this test is not working
    class _CapInherit(F.Capacitor):
        pickable = None
        inner: F.Capacitor

    module = _CapInherit()

    pick_part_recursively(module, DefaultSolver())

    assert not module.has_trait(F.has_part_picked)
    assert module.inner.has_trait(F.has_part_picked)


@pytest.mark.xfail(reason="TODO: add support for diodes")
def test_pick_led_by_colour():
    color = F.LED.Color.YELLOW
    led = F.LED()
    led.color.constrain_subset(color)
    led.current.alias_is(L.Range.from_center_rel(10 * P.milliamp, 0.1))

    solver = DefaultSolver()
    pick_part_recursively(led, solver)

    assert led.has_trait(F.has_part_picked)
    solver.update_superset_cache(led)
    assert solver.inspect_get_known_supersets(led.color).is_subset_of(
        EnumSet.from_value(color)
    )


@pytest.mark.usefixtures("setup_project_config")
@pytest.mark.xfail(reason="TODO: add support for diodes")
def test_reject_diode_for_led():
    led = F.LED()
    led.color.constrain_subset(F.LED.Color.YELLOW)
    led.current.alias_is(L.Range.from_center_rel(10 * P.milliamp, 0.1))

    diode = F.Diode()
    diode.current.alias_is(L.Range.from_center_rel(10 * P.milliamp, 0.1))

    solver = DefaultSolver()
    candidates = get_candidates(diode.get_tree(types=F.Diode), solver)
    with pytest.raises(NotCompatibleException):
        check_and_attach_candidates([(led, c) for c in candidates[diode]], solver)


@pytest.mark.usefixtures("setup_project_config")
def test_pick_error_group():
    root = L.Module()

    # Good luck finding a 10 gigafarad capacitor!
    c1 = F.Capacitor()
    c1.capacitance.alias_is(L.Range.from_center_rel(10 * P.GF, 0.1))

    c2 = F.Capacitor()
    c2.capacitance.alias_is(L.Range.from_center_rel(20 * P.GF, 0.1))

    root.add(c1)
    root.add(c2)

    solver = DefaultSolver()

    with pytest.raises(ExceptionGroup) as ex:
        pick_part_recursively(root, solver)

    assert len(ex.value.exceptions) == 1
    assert isinstance(ex.value.exceptions[0], PickError)


@pytest.mark.usefixtures("setup_project_config")
def test_pick_dependency_simple():
    class App(Module):
        r1: F.Resistor
        r2: F.Resistor

    app = App()

    solver = DefaultSolver()
    r1r = app.r1.resistance
    r2r = app.r2.resistance
    sum_lit = L.Range.from_center_rel(100 * P.kohm, 0.2)
    (r1r + r2r).constrain_subset(sum_lit)
    r1r.constrain_subset(sum_lit - r2r)
    r2r.constrain_subset(sum_lit - r1r)

    pick_part_recursively(app, solver)


@pytest.mark.usefixtures("setup_project_config")
@pytest.mark.slow
def test_pick_dependency_advanced_1():
    rdiv = F.ResistorVoltageDivider()
    rdiv.total_resistance.constrain_subset(L.Range.from_center_rel(100 * P.kohm, 0.1))
    rdiv.ratio.constrain_subset(L.Range.from_center_rel(0.1, 0.2))

    solver = DefaultSolver()
    pick_part_recursively(rdiv, solver)


@pytest.mark.usefixtures("setup_project_config")
@pytest.mark.slow
def test_pick_dependency_advanced_2():
    rdiv = F.ResistorVoltageDivider()

    rdiv.v_in.alias_is(L.Range.from_center_rel(10 * P.V, 0.1))
    rdiv.v_out.constrain_subset(L.Range(3 * P.V, 3.2 * P.V))
    rdiv.max_current.constrain_subset(L.Range(1 * P.mA, 3 * P.mA))

    solver = DefaultSolver()
    pick_part_recursively(rdiv, solver)


@pytest.mark.usefixtures("setup_project_config")
def test_null_solver():
    capacitance = L.Range.from_center_rel(10 * P.nF, 0.2)

    class App(Module):
        cap: F.Capacitor

        def __preinit__(self):
            self.cap.add(F.has_package_requirements(size=SMDSize.I0805))
            self.cap.capacitance.constrain_subset(capacitance)

    app = App()

    solver = NullSolver()
    pick_part_recursively(app, solver)

    assert app.cap.has_trait(F.has_part_picked)
    assert (
        app.cap.get_trait(F.has_package_requirements).get_sizes(solver) == SMDSize.I0805
    )
    assert (
        (solver)
        .inspect_get_known_supersets(app.cap.capacitance)
        .is_subset_of(capacitance)
    )


@pytest.mark.usefixtures("setup_project_config")
@pytest.mark.slow
def test_pick_voltage_divider_complex():
    class App(Module):
        supply: F.ElectricPower
        rdiv: F.ResistorVoltageDivider
        adc_input: F.ElectricSignal

        def __preinit__(self):
            self.supply.connect(self.rdiv.power)
            self.rdiv.output.connect(self.adc_input)

            # param
            self.supply.voltage.alias_is(L.Range(9.9 * P.V, 10.1 * P.V))
            self.adc_input.reference.voltage.constrain_subset(
                L.Range(3.0 * P.V, 3.2 * P.V)
            )
            self.rdiv.max_current.constrain_subset(L.Range(1 * P.mA, 2 * P.mA))

    app = App()
    F.is_bus_parameter.resolve_bus_parameters(app.get_graph())
    solver = DefaultSolver()

    solver.simplify_symbolically(app)

    # pick_part_recursively(app, solver)

    # for m in app.get_children_modules(types=Module):
    #    if not m.has_trait(F.has_part_picked):
    #        continue
    #    print(m.get_full_name(), m.pretty_params(solver))


@pytest.mark.usefixtures("setup_project_config")
def test_pick_capacitor_temperature_coefficient():
    # the picker backend must have access to the same enum definition for this to work
    cap = F.Capacitor()
    cap.temperature_coefficient.constrain_subset(F.Capacitor.TemperatureCoefficient.X7R)

    solver = DefaultSolver()
    pick_part_recursively(cap, solver)

    assert cap.has_trait(F.has_part_picked)



# This file is part of the faebryk project
# SPDX-License-Identifier: MIT

import logging
from abc import abstractmethod
from typing import TYPE_CHECKING

from faebryk.core.module import Module
from faebryk.core.trait import Trait

logger = logging.getLogger(__name__)

if TYPE_CHECKING:
    from faebryk.library.SurgeProtection import SurgeProtection


class can_be_surge_protected(Trait):
    @abstractmethod
    def protect(self, owner: Module) -> "SurgeProtection": ...



# This file is part of the faebryk project
# SPDX-License-Identifier: MIT

import logging
from enum import Enum, auto
from typing import TYPE_CHECKING

import faebryk.library._F as F
from faebryk.core.module import Module
from faebryk.libs.library import L
from faebryk.libs.smd import SMDSize
from faebryk.libs.units import P, Quantity

# FIXME: this has to go this way to avoid gen_F detecting a circular import
if TYPE_CHECKING:
    from faebryk.library.ElectricPower import ElectricPower

logger = logging.getLogger(__name__)


class Capacitor(Module):
    class TemperatureCoefficient(Enum):
        Y5V = auto()
        Z5U = auto()
        X7S = auto()
        X5R = auto()
        X6R = auto()
        X7R = auto()
        X8R = auto()
        C0G = auto()

    unnamed = L.list_field(2, F.Electrical)

    capacitance = L.p_field(
        units=P.F,
        likely_constrained=True,
        soft_set=L.Range(100 * P.pF, 1 * P.F),
        tolerance_guess=10 * P.percent,
    )
    # Voltage at which the design may be damaged
    max_voltage = L.p_field(
        units=P.V,
        likely_constrained=True,
        soft_set=L.Range(10 * P.V, 100 * P.V),
    )
    temperature_coefficient = L.p_field(
        domain=L.Domains.ENUM(TemperatureCoefficient),
    )

    attach_to_footprint: F.can_attach_to_footprint_symmetrically
    designator_prefix = L.f_field(F.has_designator_prefix)(
        F.has_designator_prefix.Prefix.C
    )

    @L.rt_field
    def pickable(self) -> F.is_pickable_by_type:
        return F.is_pickable_by_type(
            F.is_pickable_by_type.Type.Capacitor,
            {
                "capacitance": self.capacitance,
                "max_voltage": self.max_voltage,
                "temperature_coefficient": self.temperature_coefficient,
            },
        )

    @L.rt_field
    def can_bridge(self):
        return F.can_bridge_defined(*self.unnamed)

    @L.rt_field
    def simple_value_representation(self):
        S = F.has_simple_value_representation_based_on_params_chain.Spec
        return F.has_simple_value_representation_based_on_params_chain(
            S(self.capacitance, tolerance=True),
            S(self.max_voltage),
            S(self.temperature_coefficient),
        )

    def explicit(
        self,
        nominal_capacitance: Quantity | None = None,
        tolerance: float | None = None,
        size: SMDSize | None = None,
    ):
        if nominal_capacitance is not None:
            if tolerance is None:
                tolerance = 0.2
            capacitance = L.Range.from_center_rel(nominal_capacitance, tolerance)
            self.capacitance.constrain_subset(capacitance)

        if size is not None:
            self.add(F.has_package_requirements(size=size))

    # TODO: remove @https://github.com/atopile/atopile/issues/727
    @property
    def p1(self) -> F.Electrical:
        """One side of the capacitor."""
        return self.unnamed[0]

    @property
    def p2(self) -> F.Electrical:
        """The other side of the capacitor."""
        return self.unnamed[1]

    class _has_power(L.Trait.decless()):
        """
        This trait is used to add power interfaces to
        capacitors who use them, keeping the interfaces
        off caps which don't use it.

        Caps have power-interfaces when used with them.
        """

        def __init__(self, power: "ElectricPower") -> None:
            super().__init__()
            self.power = power

    @property
    def power(self) -> "ElectricPower":
        """An `ElectricPower` interface, which is connected to the capacitor."""
        # FIXME: this has to go this way to avoid gen_F detecting a circular import
        from faebryk.library.ElectricPower import ElectricPower

        if self.has_trait(self._has_power):
            power = self.get_trait(self._has_power).power
        else:
            power = ElectricPower()
            self.add(power, name="power_shim")
            power.hv.connect_via(self, power.lv)
            self.add(self._has_power(power))

        return power



# This file is part of the faebryk project
# SPDX-License-Identifier: MIT

from faebryk.core.moduleinterface import ModuleInterface
from faebryk.core.node import Node


class Electrical(ModuleInterface):
    """
    Electrical interface.
    """

    # potential= L.p_field(units=P.dimensionless)

    def get_net(self):
        from faebryk.library.Net import Net

        nets = {
            net
            for mif in self.get_connected()
            if (net := mif.get_parent_of_type(Net)) is not None
        }

        if not nets:
            return None

        assert len(nets) == 1
        return next(iter(nets))

    def net_crosses_pad_boundary(self) -> bool:
        from faebryk.library.Pad import Pad

        def _get_pad(n: Node):
            if (parent := n.get_parent()) is None:
                return None

            parent_node, name_on_parent = parent

            return (
                parent_node
                if isinstance(parent_node, Pad) and name_on_parent == "net"
                else None
            )

        net = self.get_connected().keys()
        pads_on_net = {pad for n in net if (pad := _get_pad(n)) is not None}

        return len(pads_on_net) > 1



# This file is part of the faebryk project
# SPDX-License-Identifier: MIT


import pytest

import faebryk.library._F as F
from faebryk.core.module import Module
from faebryk.core.solver.defaultsolver import DefaultSolver
from faebryk.libs.app.checks import check_design
from faebryk.libs.exceptions import UserDesignCheckException
from faebryk.libs.library import L


class ConfigurableI2CClient(Module):
    addressor = L.f_field(F.Addressor)(address_bits=3)
    i2c: F.I2C
    config = L.list_field(3, F.ElectricLogic)
    ref: F.ElectricPower

    @L.rt_field
    def single_electric_reference(self):
        return F.has_single_electric_reference_defined(
            F.ElectricLogic.connect_all_module_references(self)
        )

    def __preinit__(self) -> None:
        self.addressor.address.alias_is(self.i2c.address)
        self.addressor.base.alias_is(16)
        for a, b in zip(self.addressor.address_lines, self.config):
            a.connect(b)


def test_addressor():
    app = ConfigurableI2CClient()

    # app.addressor.offset.alias_is(3)
    app.i2c.address.alias_is(16 + 3)

    solver = DefaultSolver()
    solver.simplify(app)

    assert solver.inspect_get_known_supersets(app.i2c.address) == 16 + 3

    assert app.config[0].line.is_connected_to(app.ref.hv)
    assert app.config[1].line.is_connected_to(app.ref.hv)
    assert app.config[2].line.is_connected_to(app.ref.lv)


class I2CBusTopology(Module):
    server: F.I2C
    clients = L.list_field(3, ConfigurableI2CClient)

    def __init__(self, isolated=False):
        super().__init__()
        self._isolated = isolated

    def __preinit__(self) -> None:
        # self.server.address.alias_is(0)
        if not self._isolated:
            self.server.connect(*[c.i2c for c in self.clients])
            self.server.terminate(self)
        else:
            self.server.connect_shallow(*[c.i2c for c in self.clients])
            # for c in [self.server] + [c.i2c for c in self.clients]:
            #    c.terminate(self)


def test_i2c_unique_addresses():
    app = I2CBusTopology()
    app.clients[0].addressor.offset.alias_is(1)
    app.clients[1].addressor.offset.alias_is(2)
    app.clients[2].addressor.offset.alias_is(3)

    solver = DefaultSolver()
    solver.simplify(app)
    app.add(F.has_solver(solver))

    check_design(app.get_graph(), stage=F.implements_design_check.CheckStage.POST_SOLVE)


def test_i2c_duplicate_addresses():
    app = I2CBusTopology()
    app.clients[0].addressor.offset.alias_is(1)
    app.clients[1].addressor.offset.alias_is(3)
    app.clients[2].addressor.offset.alias_is(3)

    solver = DefaultSolver()
    solver.simplify(app)
    app.add(F.has_solver(solver))

    # with pytest.raises(F.I2C.requires_unique_addresses.DuplicateAddressException):
    with pytest.raises(ExceptionGroup) as e:
        check_design(
            app.get_graph(), stage=F.implements_design_check.CheckStage.POST_SOLVE
        )
    assert e.group_contains(
        UserDesignCheckException, match="Duplicate I2C addresses found on the bus:"
    )


def test_i2c_duplicate_addresses_isolated():
    app = I2CBusTopology(isolated=True)
    app.clients[0].addressor.offset.alias_is(1)
    app.clients[1].addressor.offset.alias_is(3)
    app.clients[2].addressor.offset.alias_is(3)

    solver = DefaultSolver()
    solver.simplify(app)
    app.add(F.has_solver(solver))

    # with pytest.raises(F.I2C.requires_unique_addresses.DuplicateAddressException):
    with pytest.raises(ExceptionGroup) as e:
        check_design(
            app.get_graph(), stage=F.implements_design_check.CheckStage.POST_SOLVE
        )
    assert e.group_contains(
        UserDesignCheckException, match="Duplicate I2C addresses found on the bus:"
    )



import logging
from os import PathLike
from pathlib import Path

from antlr4 import CommonTokenStream, FileStream, InputStream, Token
from antlr4.error.ErrorListener import ErrorListener

from atopile.parser.AtoLexer import AtoLexer
from atopile.parser.AtoParser import AtoParser

from .errors import UserFileNotFoundError, UserSyntaxError

log = logging.getLogger(__name__)
log.setLevel(logging.INFO)


class ErrorListenerConverter(ErrorListener):
    """Converts an error into an AtoSyntaxError."""

    def __init__(self):
        super().__init__()
        self.errors: list[UserSyntaxError] = []

    def syntaxError(
        self,
        recognizer,
        offendingSymbol: Token,
        line: int,
        column: int,
        msg: str,
        e: Exception | None,
    ):
        if e is not None and e.args != (None,):
            msg = f"{e} {msg}"

        input_stream: CommonTokenStream = recognizer.getInputStream()
        # This fill is required to get context past the offending symbol
        # to accurately report the error
        input_stream.fill()

        self.errors.append(
            UserSyntaxError.from_tokens(
                input_stream, offendingSymbol, None, msg, markdown=False
            )
        )


def make_parser(src_stream: InputStream) -> tuple[AtoParser, ErrorListenerConverter]:
    """Make a parser from a stream."""
    lexer = AtoLexer(src_stream)
    stream = CommonTokenStream(lexer)
    parser = AtoParser(stream)

    parser.removeErrorListeners()
    listener = ErrorListenerConverter()
    parser.addErrorListener(listener)

    return parser, listener


def _parse_input(
    input: InputStream, raise_multiple_errors: bool = False
) -> AtoParser.File_inputContext:
    parser, listener = make_parser(input)
    tree = parser.file_input()

    match listener.errors:
        case []:
            return tree
        case [error]:
            raise error
        case errors:
            if raise_multiple_errors:
                raise ExceptionGroup("Multiple syntax errors found", errors)

            raise errors[0]


def parse_text_as_file(
    src_code: str,
    src_path: None | str | Path = None,
    raise_multiple_errors: bool = False,
) -> AtoParser.File_inputContext:
    """Parse a string as a file input."""
    input = InputStream(src_code)
    input.name = str(src_path)
    return _parse_input(input, raise_multiple_errors=raise_multiple_errors)


def parse_file(
    src_path: str | Path, raise_multiple_errors: bool = False
) -> AtoParser.File_inputContext:
    """Parse a file from a path."""
    input = FileStream(str(src_path), encoding="utf-8")
    input.name = str(src_path)
    return _parse_input(input, raise_multiple_errors=raise_multiple_errors)


class FileParser:
    """Parses a file."""

    def __init__(self) -> None:
        self.cache = {}

    def get_ast_from_file(self, src_origin: PathLike) -> AtoParser.File_inputContext:
        """Get the AST from a file."""

        src_origin_str = str(src_origin)
        src_origin_path = Path(src_origin)

        if src_origin_str not in self.cache:
            if not src_origin_path.exists():
                raise UserFileNotFoundError(src_origin_str)
            self.cache[src_origin_str] = parse_file(src_origin_path)

        return self.cache[src_origin_str]

    def get_ast_from_text(
        self, src_code: str, src_path: Path | None = None
    ) -> AtoParser.File_inputContext:
        """Get the AST from a string."""
        # TODO: caching
        return parse_text_as_file(src_code, src_path, raise_multiple_errors=True)


parser = FileParser()



import pytest

from atopile import telemetry


@pytest.mark.parametrize(
    ("git_remote",),
    [
        ("https://github.com/atopile/atopile.git",),
        ("git@github.com:atopile/atopile.git",),
    ],
)
def test_normalize_git_remote_url(git_remote):
    assert (
        telemetry._normalize_git_remote_url(git_remote) == "github.com/atopile/atopile"
    )


def test_init_client():
    assert isinstance(telemetry.client, telemetry.Posthog)


def test_capture_event():
    @telemetry.capture(
        "test_start",
        "test_end",
        {"test_property": "test_value"},
    )
    def test_capture_event():
        pass

    test_capture_event()


def test_capture_exception():
    try:
        raise Exception("test_exception")
    except Exception as e:
        telemetry.capture_exception(e, {"test_property": "test_value"})



# This file is part of the faebryk project
# SPDX-License-Identifier: MIT

import faebryk.library._F as F
from faebryk.core.moduleinterface import ModuleInterface


class RS485HalfDuplex(ModuleInterface):
    """
    Half-duplex RS485 interface
    A = p
    B = n
    """

    diff_pair: F.DifferentialPair



# This file is part of the faebryk project
# SPDX-License-Identifier: MIT

import functools
import logging
from dataclasses import asdict, dataclass, field, make_dataclass
from textwrap import indent
from typing import Any, Iterable

from dataclasses_json import config as dataclass_json_config
from dataclasses_json import dataclass_json

import faebryk.library._F as F
from faebryk.core.module import Module
from faebryk.core.parameter import Parameter
from faebryk.libs.exceptions import UserException, downgrade
from faebryk.libs.picker.lcsc import PickedPartLCSC
from faebryk.libs.picker.lcsc import attach as lcsc_attach
from faebryk.libs.sets.sets import P_Set
from faebryk.libs.util import Serializable, SerializableJSONEncoder

logger = logging.getLogger(__name__)


@dataclass(frozen=True)
class Interval:
    min: float | None
    max: float | None


ApiParamT = P_Set | None


def SerializableField():
    return field(
        metadata=dataclass_json_config(encoder=SerializableJSONEncoder().default)
    )


# Consider moving to a more general markdown utility
def _md_list(items: Iterable[str]) -> str:
    return "\n".join(f"- {item}" for item in items)


# Consider making this a mixin instead
def _pretty_params_helper(params) -> str:
    def _map(v: Any) -> str:
        if v is None:
            return "**unconstrained**"
        elif isinstance(v, (P_Set, int, float)):
            return f"`{v}`"
        elif isinstance(v, str):
            return f'"{v}"'
        else:
            return str(v)

    return _md_list(f"`{k}`: {_map(v)}" for k, v in asdict(params).items())


@dataclass_json
@dataclass(frozen=True, kw_only=True)
class BaseParams(Serializable):
    package: ApiParamT = SerializableField()
    qty: int
    endpoint: str | None = None

    def serialize(self) -> dict:
        return self.to_dict()  # type: ignore

    def pretty_str(self) -> str:
        return _pretty_params_helper(self)


def _make_params_for_type(module_type: type[Module], endpoint: str) -> type:
    m = module_type()
    assert m.has_trait(F.is_pickable_by_type)

    pickable_trait = m.get_trait(F.is_pickable_by_type)
    parameters = pickable_trait.get_parameters()

    fields = [
        ("endpoint", str, field(default=endpoint, init=False)),
        *[(param_name, ApiParamT, SerializableField()) for param_name in parameters],
    ]

    cls = make_dataclass(
        f"{module_type.__name__}Params",
        fields,
        bases=(BaseParams,),
        frozen=True,
        kw_only=True,
    )
    return dataclass_json(cls)


ResistorParams = _make_params_for_type(F.Resistor, "resistors")
CapacitorParams = _make_params_for_type(F.Capacitor, "capacitors")
InductorParams = _make_params_for_type(F.Inductor, "inductors")
# DiodeParams = _make_params_for_type(F.Diode, "diodes")
# TVSParams = _make_params_for_type(F.TVS, "tvs")
# LEDParams = _make_params_for_type(F.LED, "leds")
# LDOParams = _make_params_for_type(F.LDO, "ldos")
# MOSFETParams = _make_params_for_type(F.MOSFET, "mosfets")


@dataclass_json
@dataclass(frozen=True)
class LCSCParams(Serializable):
    lcsc: int
    quantity: int

    def serialize(self) -> dict:
        return self.to_dict()  # type: ignore

    @classmethod
    def deserialize(cls, data: dict) -> "LCSCParams":
        return cls(**data)

    def pretty_str(self) -> str:
        return _pretty_params_helper(self)


@dataclass_json
@dataclass(frozen=True)
class ManufacturerPartParams(Serializable):
    manufacturer_name: str
    part_number: str
    quantity: int

    def serialize(self) -> dict:
        return self.to_dict()  # type: ignore

    @classmethod
    def deserialize(cls, data: dict) -> "ManufacturerPartParams":
        return cls(**data)

    def pretty_str(self) -> str:
        return _pretty_params_helper(self)


@dataclass_json
@dataclass
class ComponentPrice:
    qTo: int | None
    price: float
    qFrom: int | None


@dataclass_json
@dataclass
class Component:
    lcsc: int
    manufacturer_name: str
    part_number: str
    package: str
    datasheet_url: str
    description: str
    is_basic: int
    is_preferred: int
    stock: int
    price: list[ComponentPrice]
    attributes: dict[str, dict]  # FIXME: more specific

    @property
    def lcsc_display(self) -> str:
        return f"C{self.lcsc}"

    def get_price(self, qty: int = 1) -> float:
        """
        Get the price for qty of the component including handling fees

        For handling fees and component price classifications, see:
        https://jlcpcb.com/help/article/pcb-assembly-faqs
        """
        BASIC_HANDLING_FEE = 0
        PREFERRED_HANDLING_FEE = 0
        EXTENDED_HANDLING_FEE = 3

        if qty < 1:
            raise ValueError("Quantity must be greater than 0")

        if self.is_basic:
            handling_fee = BASIC_HANDLING_FEE
        elif self.is_preferred:
            handling_fee = PREFERRED_HANDLING_FEE
        else:
            handling_fee = EXTENDED_HANDLING_FEE

        unit_price = float("inf")
        try:
            for p in self.price:
                if p.qTo is None or qty < p.qTo:
                    unit_price = float(p.price)
            unit_price = float(self.price[-1].price)
        except LookupError:
            pass

        return unit_price * qty + handling_fee

    @functools.cached_property
    def attribute_literals(self) -> dict[str, P_Set | None]:
        def deserialize(k, v):
            if v is None:
                return None
            return P_Set.deserialize(v)

        return {k: deserialize(k, v) for k, v in self.attributes.items()}

    def attach(self, module: Module, qty: int = 1):
        lcsc_attach(module, self.lcsc_display)

        module.add(
            F.has_part_picked(
                PickedPartLCSC(
                    manufacturer=self.manufacturer_name,
                    partno=self.part_number,
                    supplier_partno=self.lcsc_display,
                    info=PickedPartLCSC.Info(
                        stock=self.stock,
                        price=self.get_price(qty),
                        description=self.description,
                        basic=bool(self.is_basic),
                        preferred=bool(self.is_preferred),
                    ),
                )
            )
        )
        module.add(F.has_datasheet_defined(self.datasheet_url))

        missing_attrs = []
        # only for type picks
        if module.has_trait(F.is_pickable_by_type):
            for name, literal in self.attribute_literals.items():
                if not hasattr(module, name):
                    missing_attrs.append(name)
                    continue

                p = getattr(module, name)
                assert isinstance(p, Parameter)
                if literal is None:
                    literal = p.domain.unbounded(p)

                p.alias_is(literal)

        if missing_attrs:
            with downgrade(UserException):
                # TODO: suggest specific library module
                # (requires Component to know its type)
                attrs = "\n".join(f"- {attr}" for attr in missing_attrs)
                raise UserException(
                    f"Module `{module}` is missing attributes:\n{attrs}\n\n"
                    "Consider using the standard library version of this module.",
                    title="Attribute(s) not found",
                )

        if logger.isEnabledFor(logging.DEBUG):
            logger.debug(
                f"Attached component {self.lcsc_display} to module {module}: \n"
                f"{indent(str(self.attributes), ' ' * 4)}\n--->\n"
                f"{indent(module.pretty_params(), ' ' * 4)}"
            )

    def __rich_repr__(self):
        yield f"{type(self).__name__}({self.lcsc_display})"

    class ParseError(Exception):
        pass



# This file is part of the faebryk project
# SPDX-License-Identifier: MIT

import logging
from abc import abstractmethod

import faebryk.library._F as F
from faebryk.core.trait import Trait

logger = logging.getLogger(__name__)


class is_surge_protected(Trait):
    @abstractmethod
    def get_protection(self) -> F.SurgeProtection: ...
